import shlex
from pyroute2 import IPRoute, NDB
from logging import getLogger
from subprocess import check_call, CalledProcessError, Popen, DEVNULL
from os.path import join
from psutil import process_iter
from tabulate import tabulate
from typing import List

from vnet_manager.conf import settings
from vnet_manager.utils.mac import random_mac_generator

logger = getLogger(__name__)


def get_vnet_interface_names_from_config(config: dict) -> List[str]:
    """
    Gets the VNet inetface names from the config
    :param dict config: The conifg generated by get_config()
    :return: list: The VNet interface names
    """
    return [settings.VNET_BRIDGE_NAME + str(i) for i in range(0, config["switches"])]


def get_machines_by_vnet_interface_name(config: dict, ifname: str) -> List[str]:
    """
    Returns a list of machine that use a particular VNet interface
    :param dict config: The config generated by get_config()
    :param str ifname: The interface to check for
    :return: list of VNet machines using that interface
    """
    machines = []
    for m_name, m_data in config["machines"].items():
        for int_data in m_data["interfaces"].values():
            if int(int_data["bridge"]) == int(ifname[-1]):
                machines.append(m_name)
    return machines


def show_vnet_interface_status(config: dict):
    """
    Shows the VNet interface status to the user
    :param dict config: The config generated by get_config()
    """
    logger.info("Listing VNet interface statuses")
    header = ["Name", "Status", "L2_addr", "Sniffer", "STP", "Used by"]
    statuses = []
    ip = IPRoute()
    ndb = NDB(log=False)
    for ifname in get_vnet_interface_names_from_config(config):
        used_by = get_machines_by_vnet_interface_name(config, ifname)
        dev = ip.link_lookup(ifname=ifname)
        if not dev:
            # Link does not exist
            statuses.append([ifname, "NA", "NA", "NA", "NA", ", ".join(used_by)])
        else:
            # Get the link info
            sniffer = check_if_sniffer_exists(ifname)
            with ndb.interfaces[ifname] as info:
                statuses.append([ifname, info["state"], info["address"], sniffer, bool(info["br_stp_state"]), ", ".join(used_by)])
    print(tabulate(statuses, headers=header, tablefmt="pretty"))


def show_vnet_veth_interface_status(config: dict):
    """
    Shows the VNet veth interface status to the user
    Assumes that the 'veths' config is present
    :param dict config: The config generated by get_config
    """
    logger.info("Listing VNet veth interface statuses")
    header = ["Name", "Status", "L2_addr", "Peer", "Master"]
    statuses = []
    ip = IPRoute()
    for name, data in config["veths"].items():
        dev = ip.link_lookup(ifname=name)
        if not dev:
            # Link does not exist
            statuses.append([name, "NA", "NA", "NA", data["bridge"]])
        else:
            # Get the link info
            info = ip.link("get", index=dev[0])
            state = info[0]["state"]
            l2_addr = [attr[1] for attr in info[0]["attrs"] if attr[0] == "IFLA_ADDRESS"][0]
            peer_id = [attr[1] for attr in info[0]["attrs"] if attr[0] == "IFLA_LINK"][0]
            peer_name = [attr[1] for attr in ip.link("get", index=peer_id)[0]["attrs"] if attr[0] == "IFLA_IFNAME"][0]
            master_id = [attr[1] for attr in info[0]["attrs"] if attr[0] == "IFLA_MASTER"][0]
            master_name = [attr[1] for attr in ip.link("get", index=master_id)[0]["attrs"] if attr[0] == "IFLA_IFNAME"][0]
            statuses.append([name, state, l2_addr, peer_name, master_name])
    print(tabulate(statuses, headers=header, tablefmt="pretty"))


def check_if_interface_exists(ifname: str) -> bool:
    """
    Check if an interface exists
    :param str ifname: The interface name to check for
    :return: bool: True if the interface exists, False otherwise
    """
    return bool(IPRoute().link_lookup(ifname=ifname))


def create_vnet_interface(ifname: str):
    """
    Creates a VNet bridge interface
    :param str ifname: The name of the interface to create
    """
    logger.info("Creating VNet bridge interface {}".format(ifname))
    IPRoute().link("add", ifname=ifname, kind="bridge")
    # Bring up the interface
    configure_vnet_interface(ifname)


def create_veth_interface(name: str, data: dict):
    """
    Creates a veth interface pair
    :param str name: The name of the veth interface to create
    :param dict data: The bridge and peer data
    """
    # We only create the interface if it has a peer
    if "peer" in data:
        IPRoute().link("add", ifname=name, kind="veth", peer=data["peer"])


def create_vnet_interface_iptables_rules(ifname: str):
    """
    VNet interfaces should act as dump bridges and should not have any connectivity to the outside world
    So this function makes some IPtables rules to make sure the VNet interface cannot talk to the outside.
    :param str ifname: The interface the create IPtables rules for
    """
    rule = "OUTPUT -o {} -j DROP".format(ifname)
    # First we check if the rule already exists
    try:
        check_call(shlex.split("iptables -C {}".format(rule)), stderr=DEVNULL)
        logger.debug("IPtables DROP rule for VNet interface {} already exists, skipping creation".format(ifname))
    except CalledProcessError:
        logger.info("Creating IPtables DROP rule to the outside world for VNet interface {}".format(ifname))
        try:
            check_call(shlex.split("iptables -A {}".format(rule)))
        except CalledProcessError as e:
            logger.error("Unable to create IPtables rule, got output: {}".format(e.output))


def configure_vnet_interface(ifname: str):
    """
    Configures an vnet interface to be in the correct state for forwarding vnet machine traffic
    :param str ifname: The vnet interface to configure
    """
    ip = IPRoute()
    dev = ip.link_lookup(ifname=ifname)[0]
    # Make sure it's set to down state
    ip.link("set", index=dev, state="down")
    # Set the mac
    ip.link("set", index=dev, address=random_mac_generator())
    # Finally, bring up the interface
    ip.link("set", index=dev, state="up")


def configure_veth_interface(name: str, data: dict):
    """
    Configures a veth interface, connects to the correct bridge
    :param str name: The name of the veth interface
    :param dict data: The veth interface data (bridge name)
    """
    logger.info("Creating VNet veth interface {}".format(name))
    ip = IPRoute()
    dev = ip.link_lookup(ifname=name)[0]
    bridge = ip.link_lookup(ifname=data["bridge"])[0]
    # Connect the veth interface to the bridge
    ip.link("set", index=dev, master=bridge)


def bring_up_vnet_interfaces(config: dict, sniffer: bool = False):
    """
    Check the status of the vnet interfaces defined in the config and brings up the interfaces if needed
    :param dict config: The config generated by get_config()
    :param bool sniffer: Check for a sniffer process and create it if it does not exist
    """
    ip = IPRoute()
    for ifname in get_vnet_interface_names_from_config(config):
        if not check_if_interface_exists(ifname):
            create_vnet_interface(ifname)
        # Block traffic to the outside world
        create_vnet_interface_iptables_rules(ifname)
        # Make sure the interface is up
        ip.link("set", ifname=ifname, state="up")
        if sniffer and not check_if_sniffer_exists(ifname):
            # Create it
            start_tcpdump_on_vnet_interface(ifname)
    if "veths" in config:
        ensure_vnet_veth_interfaces(config)


def ensure_vnet_veth_interfaces(config: dict):
    """
    Create en configure the veth interfaces defined in the VNet config
    Assumes there are veth interfaces present in the config
    :param dict config: The config generated by get_config()
    """
    logger.info("VNet veth config found, ensuring interfaces")
    for name, data in config["veths"].items():
        # Set STP on the master if required
        if "stp" in data:
            logger.info("{} STP on VNet interface {}".format("Enabling" if data["stp"] else "Disabling", data["bridge"]))
            state = 1 if data["stp"] else 0
            nbd = NDB(log=False)
            with nbd.interfaces[data["bridge"]] as bridge:
                bridge.set("br_stp_state", state)
        if not check_if_interface_exists(name):
            create_veth_interface(name, data)
        # Always configure a VNet veth interface to make sure it is connected to its master bridge
        configure_veth_interface(name, data)
        configure_vnet_interface(name)


def check_if_sniffer_exists(ifname: str) -> bool:
    """
    Check if there is already a sniffer running for a VNet interface
    :param str ifname: The VNet interface name to check
    :return bool: True if it exists, False otherwise
    """
    for process in process_iter():
        process_line = process.cmdline()
        if "tcpdump" in process_line and ifname in process_line:
            logger.debug("A TCPdump sniffer for interface {} already exists".format(ifname))
            return True
    return False


def bring_down_vnet_interfaces(config: dict):
    """
    Brings down the VNet interfaces defined in the config
    This will automatically kill any attached sniffer processes
    :param dict config: The config generated by get_config()
    """
    ip = IPRoute()
    if "veths" in config:
        for name in config["veths"].keys():
            if check_if_interface_exists(name):
                logger.info("Bringing down VNet veth interface {}".format(name))
                ip.link("set", ifname=name, state="down")
    for ifname in get_vnet_interface_names_from_config(config):
        # Set the interface to down status
        if check_if_interface_exists(ifname):
            logger.info("Bringing down VNet interface {}".format(ifname))
            ip.link("set", ifname=ifname, state="down")
        else:
            # Device doesn't exist
            logger.warning("Tried to bring down VNet interface {}, but the interface doesn't exist".format(ifname))


def delete_vnet_interfaces(config: dict):
    """
    Delete the VNet interfaces defined in the config
    :param config:
    """
    ip = IPRoute()
    if "veths" in config:
        for name, data in config["veths"].items():
            # Veth interfaces are deleted in pairs, so we only delete the ones with a peer
            if "peer" in data and check_if_interface_exists(name):
                logger.info("Deleting VNet veth interface {}".format(name))
                ip.link("del", ifname=name)
    for ifname in get_vnet_interface_names_from_config(config):
        # Delete the interface
        if check_if_interface_exists(ifname):
            logger.info("Deleting VNet interface {}".format(ifname))
            ip.link("del", ifname=ifname)
        else:
            # Device doesn't exist
            logger.info("Tried to delete VNet interface {}, but it is already gone. That's okay".format(ifname))


def start_tcpdump_on_vnet_interface(ifname: str):
    """
    Starts a tcpdump process on a vnet interface
    :param str ifname: The interface to start the tcpdump on
    """
    path = join(settings.VNET_SNIFFER_PCAP_DIR, "{}.pcap".format(ifname))
    logger.info("Starting sniffer on VNet interface {}, PCAP location: {}".format(ifname, path))
    Popen(shlex.split("tcpdump -i {} -U -w {}".format(ifname, path)))



import logging
from argparse import ArgumentParser

from vnet_manager.log import setup_console_logging, get_logging_verbosity
from vnet_manager.tests import VNetTestCase
from vnet_manager.conf import settings


def my_test_args(args):
    parser = ArgumentParser()
    parser.add_argument("-q", "--quite", action="count", default=0)
    parser.add_argument("-v", "--verbose", action="count", default=0)
    return parser.parse_args(args=args)


class TestGetLoggingVerbosity(VNetTestCase):
    def test_get_logging_verbosity_never_returns_higher_number_then_defined_verbs(self):
        args = my_test_args(["-vvvvvvvvvvv", "--verbose", "-v"])
        self.assertEqual(get_logging_verbosity(args), logging.DEBUG)

    def test_get_logging_verbosity_never_returns_lower_number_then_defined_verbs(self):
        args = my_test_args(["-qqqqqqqqqqq", "--quite", "-q"])
        self.assertEqual(get_logging_verbosity(args), logging.CRITICAL)

    def test_get_logging_verbosity_return_error_level_if_two_quites_passed(self):
        args = my_test_args(["--quite", "-q"])
        self.assertEqual(get_logging_verbosity(args), logging.ERROR)


class TestLog(VNetTestCase):
    def test_setup_console_logging_sets_up_INFO_logging_by_default(self):
        setup_console_logging()
        self.assertEqual(settings.LOGGING["handlers"]["console"]["level"], logging.INFO)

    def test_setup_console_logging_sets_up_passed_logging_level(self):
        setup_console_logging(logging.DEBUG)
        self.assertEqual(settings.LOGGING["handlers"]["console"]["level"], logging.DEBUG)



from logging import getLogger

from vnet_manager.providers.lxc import get_lxd_client
from vnet_manager.conf import settings

logger = getLogger(__name__)


def check_if_lxc_profile_exists(name: str) -> bool:
    """
    Check if an LXC profile exists
    :param str name: The LXC profile to check for
    :return bool: True if it exists, false otherwise
    """
    return get_lxd_client().profiles.exists(name)


def create_vnet_lxc_profile(name: str):
    """
    Create a VNet specific LXC profile
    :param str name: The LXC profile to create
    """
    client = get_lxd_client()
    # Check if the profile already exists
    if check_if_lxc_profile_exists(name):
        raise RuntimeError("Tried to create VNet LXC profile {}, but it already exists".format(name))

    devices = {
        # Disk config
        "root": {
            "path": "/",
            "pool": settings.LXC_STORAGE_POOL_NAME,
            "type": "disk",
        }
    }
    logger.info("Creating LXC profile for storage pool {}".format(settings.LXC_STORAGE_POOL_NAME))
    client.profiles.create(name, config={}, devices=devices)


def delete_vnet_lxc_profile(name: str):
    """
    Deletes a VNet specific LXC profile
    Profile must not be used when deleting
    :param str name: The LXC profile to delete
    """
    # Check if the profile even exists
    if not check_if_lxc_profile_exists(name):
        logger.warning("Tried to delete LXC profile {}, but it didn't exist, skipping...".format(name))
        return

    client = get_lxd_client()
    # Check if the profile is still in use
    profile = client.profiles.get(name)
    if profile.used_by:
        logger.error("LXC profile {} still used by: {}".format(name, ", ".join(profile.used_by)))
        logger.critical("LXC profile {} is still in use, please destroy any remaining machines first".format(name))
        raise RuntimeError("LXC profile {} is still in use".format(name))
    logger.info("Deleting LXC profile {}".format(name))
    profile.delete()



from copy import deepcopy
from unittest.mock import call, patch, mock_open, MagicMock
from os.path import join
from pylxd.exceptions import NotFound

from vnet_manager.tests import VNetTestCase
from vnet_manager.operations.files import (
    put_files_on_machine,
    select_files_and_put_on_machine,
    place_file_on_lxc_machine,
    generate_vnet_hosts_file,
    place_vnet_hosts_file_on_machines,
    write_file_to_lxc_container,
)
from vnet_manager.conf import settings


class TestPutFilesOnMachine(VNetTestCase):
    def setUp(self) -> None:
        # Use VALIDATED_CONFIG because here the files part has been expanded to the absolute paths
        self.config = deepcopy(settings.VALIDATED_CONFIG)
        self.select_and_put_on_machine = self.set_up_patch("vnet_manager.operations.files.select_files_and_put_on_machine")

    def test_put_files_on_machine_gets_file_paths_from_config_and_calls_select_files_and_put_on_machine(self):
        calls = [
            call(name, data["files"], settings.MACHINE_TYPE_PROVIDER_MAPPING[data["type"]])
            for name, data in self.config["machines"].items()
            if "files" in data
        ]
        put_files_on_machine(self.config)
        self.select_and_put_on_machine.assert_has_calls(calls)

    def test_put_files_on_machine_does_not_call_select_files_and_put_on_machine_when_there_are_no_machines_with_files(self):
        del self.config["machines"]["router100"]
        del self.config["machines"]["router101"]
        put_files_on_machine(self.config)
        self.assertFalse(self.select_and_put_on_machine.called)


class TestSelectFilesAndPutOnMachine(VNetTestCase):
    def setUp(self) -> None:
        self.logger = self.set_up_patch("vnet_manager.operations.files.logger")
        self.is_dir = self.set_up_patch("vnet_manager.operations.files.isdir")
        self.is_dir.return_value = True
        self.is_file = self.set_up_patch("vnet_manager.operations.files.isfile")
        self.is_file.return_value = True
        self.place_file_on_lxc_machine = self.set_up_patch("vnet_manager.operations.files.place_file_on_lxc_machine")
        self.list_dir = self.set_up_patch("vnet_manager.operations.files.listdir")
        self.list_dir.return_value = ["file1", "file2", "file3"]
        self.machine = "router100"
        self.files = settings.VALIDATED_CONFIG["machines"]["router100"]["files"]

    def test_select_files_and_put_on_machine_calls_check_methods_to_check_for_file_or_dir(self):
        self.is_dir.return_value = False
        select_files_and_put_on_machine(self.machine, self.files, "lxc")
        self.is_dir.assert_called_once_with(next(iter(self.files)))
        self.is_file.assert_called_once_with(next(iter(self.files)))

    def test_select_files_and_put_on_machine_put_a_single_file_on_machine(self):
        self.is_dir.return_value = False
        select_files_and_put_on_machine(self.machine, self.files, "lxc")
        self.place_file_on_lxc_machine.assert_called_once_with(self.machine, next(iter(self.files)), next(iter(self.files.values())))

    def test_select_files_and_put_on_machine_calls_listdir_function(self):
        select_files_and_put_on_machine(self.machine, self.files, "lxc")
        self.list_dir.assert_called_once_with(next(iter(self.files)))

    def test_select_files_and_put_on_machine_calls_place_method_for_each_file_in_dir(self):
        select_files_and_put_on_machine(self.machine, self.files, "lxc")
        calls = [
            call(self.machine, join(next(iter(self.files)), file), join(next(iter(self.files.values())), file))
            for file in self.list_dir.return_value
        ]
        self.place_file_on_lxc_machine.assert_has_calls(calls)

    def test_select_files_and_put_on_machine_logs_error_if_file_is_not_a_dir_or_file(self):
        self.is_dir.return_value = False
        self.is_file.return_value = False
        select_files_and_put_on_machine(self.machine, self.files, "lxc")
        self.logger.error.assert_called_once_with(
            "Tried to select file {} for copying, but it is neither a file nor a directory".format(next(iter(self.files)))
        )


class TestPlaceFileOnLXCContainer(VNetTestCase):
    def setUp(self) -> None:
        self.get_lxd_client = self.set_up_patch("vnet_manager.operations.files.get_lxd_client")
        self.machine = MagicMock()
        self.get_lxd_client.return_value.containers.get.return_value = self.machine
        self.is_file = self.set_up_patch("vnet_manager.operations.files.isfile")
        self.host_file_p = "/root/host"
        self.guest_file_p = "/root/guest"

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_place_file_on_lxc_machine_calls_lxd_client(self, _):
        place_file_on_lxc_machine("router100", self.host_file_p, self.guest_file_p)
        self.get_lxd_client.assert_called_once_with()

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_place_file_on_lxc_machine_calls_open(self, open_mock):
        place_file_on_lxc_machine("router100", self.host_file_p, self.guest_file_p)
        open_mock.assert_called_once_with(self.host_file_p, "r")

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_place_file_on_lxc_machine_calls_open_read_function(self, open_mock):
        place_file_on_lxc_machine("router100", self.host_file_p, self.guest_file_p)
        handle = open_mock()
        handle.read.assert_called_once_with()

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_place_file_on_lxc_machine_does_nothing_if_container_not_found(self, _):
        self.get_lxd_client.side_effect = NotFound(response=b"response")
        place_file_on_lxc_machine("router100", self.host_file_p, self.guest_file_p)
        self.assertFalse(self.machine.files.put.called)

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_place_file_on_lxc_machine_does_nothing_if_file_is_not_a_file(self, open_mock):
        self.is_file.return_value = False
        place_file_on_lxc_machine("router100", self.host_file_p, self.guest_file_p)
        self.assertFalse(open_mock.called)

    @patch("builtins.open", new_callable=mock_open, read_data="data")
    def test_place_file_on_lxc_machine_calls_files_put_method_on_machine(self, _):
        place_file_on_lxc_machine("router100", self.host_file_p, self.guest_file_p)
        self.machine.files.put.assert_called_once_with(self.guest_file_p, "data")


class TestWriteFilesToLXCContainer(VNetTestCase):
    def setUp(self) -> None:
        self.get_lxd_client = self.set_up_patch("vnet_manager.operations.files.get_lxd_client")
        self.machine = MagicMock()
        self.get_lxd_client.return_value.containers.get.return_value = self.machine

    def test_write_files_to_lxc_container_calls_lxd_client(self):
        write_file_to_lxc_container("blaap", "blaap", "blaap")
        self.get_lxd_client.assert_called_once_with()

    def test_write_files_to_lxc_container_calls_containers_get(self):
        write_file_to_lxc_container("machine", "blaap", "blaap")
        self.get_lxd_client.return_value.containers.get.assert_called_once_with("machine")

    def test_write_files_to_lxc_container_calls_files_put(self):
        write_file_to_lxc_container("blaap", "path", "data")
        self.machine.files.put.assert_called_once_with("path", "data")

    def test_write_files_to_lxc_container_does_not_call_files_put_when_container_does_not_exist(self):
        self.get_lxd_client.return_value.containers.get.side_effect = NotFound(response="banaan")
        write_file_to_lxc_container("blaap", "blaap", "blaap")
        self.assertFalse(self.machine.files.put.called)


class TestGenerateVNetHostsFile(VNetTestCase):
    def setUp(self) -> None:
        self.config = deepcopy(settings.CONFIG)
        self.excepted_hosts_file = (
            settings.VNET_STATIC_HOSTS_FILE_PART
            + "192.168.0.2   router100\nfd00:12::2   router100\n192.168.0.1   router101\nfd00:12::1   router101\n10.0.0.1   "
            "router101\nfd00:23::1   router101\n10.0.0.2   host102\nfd00:23::2   host102"
        )

    @patch("builtins.open", new_callable=mock_open)
    def test_generate_vnet_hosts_file_calls_open(self, open_mock):
        generate_vnet_hosts_file(self.config)
        open_mock.assert_called_once_with(settings.VNET_ETC_HOSTS_FILE_PATH, "w")

    @patch("builtins.open", new_callable=mock_open)
    def test_generate_vnet_hosts_file_generates_correct_hosts_file(self, open_mock):
        generate_vnet_hosts_file(self.config)
        handle = open_mock()
        handle.write.assert_called_once_with(self.excepted_hosts_file)


class TestPlaceVNetHostsFileOnMachines(VNetTestCase):
    def setUp(self) -> None:
        self.select_and_put = self.set_up_patch("vnet_manager.operations.files.select_files_and_put_on_machine")
        self.config = deepcopy(settings.CONFIG)

    def test_place_vnet_hosts_file_on_machines_calls_select_and_put_for_each_machine(self):
        calls = [call(name, {settings.VNET_ETC_HOSTS_FILE_PATH: "/etc/hosts"}, "lxc") for name in self.config["machines"]]
        place_vnet_hosts_file_on_machines(self.config)
        self.select_and_put.assert_has_calls(calls)



from ipaddress import IPv4Interface, IPv6Interface, ip_interface
from re import fullmatch
from logging import getLogger
from os.path import isdir, isfile, join
from copy import deepcopy

from vnet_manager.utils.mac import random_mac_generator
from vnet_manager.conf import settings

logger = getLogger(__name__)


class ValidateConfig:
    """
    Validates the config generated by get_config() and updates some values if missing
    """

    def __init__(self, config: dict):
        """
        :param dict config: The config generated by get_config()
        """
        self._all_ok = True
        self._validators_ran = 0
        self._new_config = deepcopy(config)
        self.default_message = ". Please check your settings"
        self.config = config

    def __str__(self) -> str:
        return "VNet config validator, current_state: {}, amount of validators run: {}".format(
            "OK" if self._all_ok else "NOT OK", self._validators_ran
        )

    @property
    def config_validation_successful(self) -> bool:
        """
        This property can be called to see if any unrecoverable errors in the config have been found
        """
        return self._all_ok

    @property
    def updated_config(self) -> dict:
        """
        This property contains a updated config dict, with all values that have been fixed by this validator
        """
        return self._new_config

    @property
    def validators_ran(self) -> int:
        """
        Return the amount of validators that have been run
        """
        return self._validators_ran

    def validate(self):
        """
        Run all validation functions
        """
        self._all_ok = True
        self.validate_switch_config()
        self.validate_machine_config()
        if "veths" in self.config:
            self.validate_veth_config()

    def validate_switch_config(self):
        """
        Validates the switch part of the config
        """
        self._validators_ran += 1
        if "switches" not in self.config:
            logger.error("Config item 'switches' missing{}".format(self.default_message))
            self._all_ok = False
        elif not isinstance(self.config["switches"], int):
            logger.error(
                "Config item 'switches: {}' does not seem to be an integer{}".format(self.config["switches"], self.default_message)
            )
            self._all_ok = False

    def validate_machine_config(self):
        # TODO: Refactor
        # pylint: disable=too-many-branches
        """
        Validates the machines part of the config
        """
        self._validators_ran += 1
        if "machines" not in self.config:
            logger.error("Config item 'machines' missing{}".format(self.default_message))
            self._all_ok = False
        elif not isinstance(self.config["machines"], dict):
            logger.error("Machines config is not a dict, this means the user config is incorrect{}".format(self.default_message))
            self._all_ok = False
        else:
            for name, values in self.config["machines"].items():
                if "type" not in values:
                    logger.error("Type not found for machine {}{}".format(name, self.default_message))
                    self._all_ok = False
                elif values["type"] not in settings.SUPPORTED_MACHINE_TYPES:
                    logger.error(
                        "Type {} for machine {} unsupported. I only support the following types: {}{}".format(
                            values["type"], name, settings.SUPPORTED_MACHINE_TYPES, self.default_message
                        )
                    )
                    self._all_ok = False

                # Files
                if "files" in values:
                    if not isinstance(values["files"], dict):
                        logger.error("Files directive for machine {} is not a dict{}".format(name, self.default_message))
                        self._all_ok = False
                    else:
                        # Check the files
                        self.validate_machine_files_parameters(name)

                # Interfaces
                if "interfaces" not in values:
                    logger.error("Machine {} does not appear to have any interfaces{}".format(name, self.default_message))
                    self._all_ok = False
                elif not isinstance(values["interfaces"], dict):
                    logger.error(
                        "The interfaces for machine {} are not given as a dict, this usually means a typo in the config{}".format(
                            name, self.default_message
                        )
                    )
                    self._all_ok = False
                else:
                    self.validate_interface_config(name)

                # VLANs?
                if "vlans" not in values:
                    logger.debug("Machine {} does not appear to have any VLAN interfaces, that's okay".format(name))
                elif not isinstance(values["vlans"], dict):
                    logger.error(
                        "Machine {} has a VLAN config but it does not "
                        "appear to be a dict, this usually means a typo in the config{}".format(name, self.default_message)
                    )
                    self._all_ok = False
                else:
                    self.validate_vlan_config(name)

                # Bridges?
                if "bridges" not in values:
                    logger.debug("Machine {} does not appear to have any Bridge interfaces, that's okay".format(name))
                elif not isinstance(values["bridges"], dict):
                    logger.error(
                        "Machine {} has a bridge config defined, but it is not a dictionary, "
                        "this usally means a typo in the config{}".format(name, self.default_message)
                    )
                    self._all_ok = False
                else:
                    self.validate_machine_bridge_config(name)

    def validate_vlan_config(self, machine):
        """
        Validates the VLAN config of a particular machine
        :param machine: str: the machine to validate the VLAN config for
        """
        vlans = self.config["machines"][machine]["vlans"]
        for name, values in vlans.items():
            if "id" not in values:
                logger.error("VLAN {} on machine {} is missing it's vlan id{}".format(name, machine, self.default_message))
                self._all_ok = False
            else:
                try:
                    self._new_config["machines"][machine]["vlans"][name]["id"] = int(values["id"])
                except ValueError:
                    logger.error(
                        "Unable to cast VLAN {} with ID {} from machine {} to a integer{}".format(
                            name, values["id"], machine, self.default_message
                        )
                    )
                    self._all_ok = False
            if "link" not in values:
                logger.error("VLAN {} on machine {} is missing it's link attribute{}".format(name, machine, self.default_message))
                self._all_ok = False
            elif not isinstance(values["link"], str):
                logger.error(
                    "Link {} for VLAN {} on machine {}, does not seem to be a string{}".format(
                        values["link"], name, machine, self.default_message
                    )
                )
                self._all_ok = False
            # This check requires a valid interface config, so we only do it if the previous checks have been successful
            elif self._all_ok and values["link"] not in self.config["machines"][machine]["interfaces"]:
                logger.error(
                    "Link {} for VLAN {} on machine {} does not correspond to any interfaces on the same machine{}".format(
                        values["link"], name, machine, self.default_message
                    )
                )
                self._all_ok = False
            if "addresses" not in values:
                logger.debug("VLAN {} on machine {} does not have any addresses, that's okay".format(name, machine))
            elif not isinstance(values["addresses"], list):
                logger.error(
                    "Addresses on VLAN {} for machine {}, does not seem to be a list{}".format(name, machine, self.default_message)
                )
                self._all_ok = False
            else:
                for address in values["addresses"]:
                    try:
                        ip_interface(address)
                    except ValueError as e:
                        logger.error(
                            "Address {} for VLAN {} on machine {} does not seem to be a valid address, got parse error {}".format(
                                address, name, machine, e
                            )
                        )
                        self._all_ok = False

    def validate_machine_files_parameters(self, machine: str):
        """
        Validates the files config of a particular machine
        Assumes the files dict exists for that machine
        :param str machine: The machine to validates the files config for
        """
        files = self.config["machines"][machine]["files"]
        for host_file in files.keys():
            # First check if the user gave a relative dir from the config dir
            if isdir(join(self.config["config_dir"], host_file)) or isfile(join(self.config["config_dir"], host_file)):
                logger.debug(
                    "Updating relative host_file path {} to full path {}".format(host_file, join(self.config["config_dir"], host_file))
                )
                self._new_config["machines"][machine]["files"][join(self.config["config_dir"], host_file)] = self._new_config["machines"][
                    machine
                ]["files"].pop(host_file)
            # Check for absolute paths
            elif not isdir(host_file) or not isfile(host_file):
                logger.error(
                    "Host file {} for machine {} does not seem to be a dir or a file{}".format(host_file, machine, self.default_message)
                )
                self._all_ok = False

    def validate_interface_config(self, machine: str):
        """
        Validates the interface config of a particular machine
        Assumes the interfaces dict exists for that machine
        :param str machine: the machine to validate the interfaces config for
        """
        interfaces = self.config["machines"][machine]["interfaces"]
        for int_name, int_vals in interfaces.items():
            if "ipv4" not in int_vals:
                logger.debug(
                    "No IPv4 found for interface {} on machine {}. That's okay, no IPv4 will be configured".format(int_name, machine)
                )
            else:
                # Validate the given IP
                try:
                    IPv4Interface(int_vals["ipv4"])
                except ValueError as e:
                    logger.error("Unable to parse IPv4 address {} for machine {}. Parse error: {}".format(int_vals["ipv4"], machine, e))
                    self._all_ok = False
            if "ipv6" not in int_vals:
                logger.debug(
                    "No IPv6 found for interface {} on machine {}, that's okay no IPv6 address will be configured".format(int_name, machine)
                )
            else:
                # Validate the given IP
                try:
                    IPv6Interface(int_vals["ipv6"])
                except ValueError as e:
                    logger.error("Unable to parse IPv6 address {} for machine {}. Parse error: {}".format(int_vals["ipv6"], machine, e))
                    self._all_ok = False
            if "mac" not in int_vals:
                logger.debug("MAC not found for interface {} on machine {}, generating a random one".format(int_name, machine))
                self._new_config["machines"][machine]["interfaces"][int_name]["mac"] = random_mac_generator()
            # From: https://stackoverflow.com/a/7629690/8632038
            elif not fullmatch(r"^([0-9A-Fa-f]{2}[:-]){5}([0-9A-Fa-f]{2})$", int_vals["mac"]):
                logger.error(
                    "MAC {} for interface {} on machine {}, does not seem to be valid{}".format(
                        int_vals["mac"], int_name, machine, self.default_message
                    )
                )
                self._all_ok = False
            if "bridge" not in int_vals:
                logger.error("bridge keyword missing on interface {} for machine {}{}".format(int_name, machine, self.default_message))
                self._all_ok = False
            elif not isinstance(int_vals["bridge"], int) or int_vals["bridge"] > self.config["switches"] - 1:
                logger.error(
                    "Invalid bridge number detected for interface {} on machine {}. "
                    "The bridge keyword should correspond to the interface number of the vnet bridge to connect to "
                    "(starting at iface number 0)".format(int_name, machine)
                )
                self._all_ok = False

    def validate_machine_bridge_config(self, machine: str):
        bridges = self.config["machines"][machine]["bridges"]
        for br_name, br_vals in bridges.items():
            if "ipv4" not in br_vals:
                logger.debug("Bridge {} on machine {} has no IPv4 assigned, that's okay".format(br_name, machine))
            else:
                # Validate the given IP
                try:
                    IPv4Interface(br_vals["ipv4"])
                except ValueError as e:
                    logger.error("Unable to parse IPv4 address for bridge {} on machine {}, got error: {}".format(br_name, machine, e))
                    self._all_ok = False
            if "ipv6" not in br_vals:
                logger.debug("Bridge {} on machine {} has no IPv6 address, that's okay".format(br_name, machine))
            else:
                try:
                    # Validate the IPv6 address
                    IPv6Interface(br_vals["ipv6"])
                except ValueError as e:
                    logger.error("Unable to parse IPv6 address for bridge {} on machine {}, got error: {}".format(br_name, machine, e))
                    self._all_ok = False
            if "slaves" not in br_vals:
                logger.error("Bridge {} on machine {} does not have any slaves".format(br_name, machine))
                self._all_ok = False
            elif not isinstance(br_vals["slaves"], list):
                logger.error("Slaves on bridge {} for machine {}, is not formatted as a list".format(br_name, machine))
                self._all_ok = False
            else:
                # For each slave, check if the interface exists
                for slave in br_vals["slaves"]:
                    if slave not in self.config["machines"][machine]["interfaces"].keys():
                        logger.error("Undefined slave interface {} assigned to bridge {} on machine {}".format(slave, br_name, machine))
                        self._all_ok = False

    def validate_veth_config(self):
        """
        Validates the veth config if present
        """
        if "veths" not in self.config:
            logger.warning("Tried to validate veth config, but no veth config present, skipping...")
            return
        if not isinstance(self.config["veths"], dict):
            logger.error("Config item: 'veths' does not seem to be a dict {}".format(self.default_message))
            self._all_ok = False
            return
        for name, values in self.config["veths"].items():
            if not isinstance(name, str):
                logger.error("veth interface name: {} does not seem to be a string{}".format(name, self.default_message))
                self._all_ok = False
            elif not isinstance(values, dict):
                logger.error("veth interface {} data does not seem to be a dict{}".format(name, self.default_message))
                self._all_ok = False
            else:
                if "bridge" not in values:
                    logger.error("veth interface {} is missing the bridge parameter{}".format(name, self.default_message))
                    self._all_ok = False
                elif not isinstance(values["bridge"], str):
                    logger.error("veth interface {} bridge parameter does not seem to be a str{}".format(name, self.default_message))
                    self._all_ok = False
                if "peer" not in values:
                    logger.debug("veth interface {} does not have a peer, that's ok, assuming it's peer is defined elsewhere".format(name))
                elif not isinstance(values["peer"], str):
                    logger.error("veth interface {} peer parameter does not seem to be a string{}".format(name, self.default_message))
                    self._all_ok = False
                if "stp" not in values:
                    logger.debug("veth interface {} as no STP parameter, that's okay".format(name))
                elif not isinstance(values["stp"], bool):
                    logger.error("veth interface {} stp parameter does not seem to be a boolean{}".format(name, self.default_message))
                    self._all_ok = False



import sys
from logging import getLogger
from os import EX_NOPERM, environ
from typing import Sequence

from vnet_manager.conf import settings
from vnet_manager.log import setup_console_logging, get_logging_verbosity
from vnet_manager.actions.manager import ActionManager
from vnet_manager.utils.user import check_for_root_user
from vnet_manager.argeparser import parse_vnet_args

logger = getLogger(__name__)


def main(args: Sequence = None) -> int:
    """
    Program entry point
    :param list args: The pre-cooked arguments to pass to the ArgParser
    :return int: exit_code
    """
    args = parse_vnet_args(args)
    # Set the VNET_FORCE variable, if --yes is given this will answer yes to all questions
    environ[settings.VNET_FORCE_ENV_VAR] = "true" if args.yes else "false"
    # Setup logging
    setup_console_logging(verbosity=get_logging_verbosity(args))
    # Most VNet operation require root. So, do a root check
    if not check_for_root_user():
        logger.critical("This program should only be run as root")
        return EX_NOPERM
    # Let the action manager handle the rest
    manager = ActionManager(config_path=args.config, sniffer=args.sniffer, base_image=args.base_image, no_hosts=args.no_hosts)
    if args.machines:
        manager.machines = args.machines
    return manager.execute(args.action)


if __name__ == "__main__":
    sys.exit(main())



from unittest.mock import Mock

from vnet_manager.conf import Settings
from vnet_manager.tests import VNetTestCase


class TestSettings(VNetTestCase):
    def setUp(self) -> None:
        self.mod = Mock(lower=123, UPPER=456, MiXeD=789)
        self.mock_importlib = self.set_up_patch("importlib.import_module", return_value=self.mod)
        self.mock_settings = Settings("module.class.sub")

    def test_that_init_only_sets_upper_variables(self):
        self.assertListEqual(["UPPER"], self.mock_settings.__iter__())

    def test_that_getitem_returns_value_of_setting(self):
        self.assertEqual(self.mock_settings["UPPER"], 456)

    def test_that_iter_returns_iterable(self):
        self.assertIsInstance(self.mock_settings.__iter__(), list)

    def test_that_len_returns_correct_length_of_all_settings(self):
        self.assertEqual(1, len(self.mock_settings))



from logging import getLogger
from pylxd.exceptions import LXDAPIException

from vnet_manager.providers.lxc import get_lxd_client
from vnet_manager.conf import settings

logger = getLogger(__name__)


def check_if_lxc_storage_pool_exists(name: str = settings.LXC_STORAGE_POOL_NAME) -> bool:
    """
    Check for the existence of the LXC storage pool defined in the settings
    :param str name: The name of the storage pool to check for (default from the settings)
    :return: bool: True if the pool exists, False otherwise
    """
    logger.debug("Checking for the existence of LXC {} storage pool".format(name))
    return get_lxd_client().storage_pools.exists(name)


def create_lxc_storage_pool(name: str = settings.LXC_STORAGE_POOL_NAME, driver: str = settings.LXC_STORAGE_POOL_DRIVER):
    """
    Creates a new LXC storage pool with the passed driver and name
    :param str name: The name of the storage pool to create (default from settings)
    :param str driver: The driver to use (default from settings)
    """
    # Sanity check, we don't want to create duplicate pools
    if check_if_lxc_storage_pool_exists(name):
        raise RuntimeError("LXC storage pool {} already exists, cannot create duplicate".format(name))

    # Make it
    logger.info("Creating LXC storage pool {} with driver {}".format(name, driver))
    client = get_lxd_client()
    try:
        client.storage_pools.create({"name": name, "driver": driver, "config": {"size": settings.LXC_STORAGE_POOL_SIZE}})
        logger.info("Storage pool {} with driver {} successfully created".format(name, driver))
    except LXDAPIException as e:
        logger.critical("Got API error while creating storage pool {}. Error: {}".format(name, e))
        raise RuntimeError("Received API error while creating storage pool {}".format(name)) from e


def delete_lxc_storage_pool(name: str):
    """
    Deletes a LXC storage pool
    Pool must be empty before deletion
    :param str name: The name of the pool to delete
    """
    # Check if the pool even exists
    if not check_if_lxc_storage_pool_exists(name):
        logger.warning("Tried to delete LXC storage pool {}, but it didn't exist, skipping...".format(name))
        return

    client = get_lxd_client()
    # Try to delete it
    try:
        logger.info("Deleting LXC storage pool {}".format(name))
        client.storage_pools.get(name).delete()
    except LXDAPIException as e:
        logger.critical("Got API error while deleting storage pool {}. Error: {}".format(name, e))
        raise RuntimeError("Received API error while deleting storage pool {}".format(name)) from e



import sys
from os import geteuid, getenv
from logging import getLogger
from typing import Any, AnyStr, Callable, Iterable

from vnet_manager.conf import settings

logger = getLogger(__name__)


def check_for_root_user() -> bool:
    """
    Checks if the user running this is root
    :return: bool: True if the user is root, False otherwise
    """
    return geteuid() == 0


def request_confirmation(
    message: AnyStr = None,
    prompt: AnyStr = "Continue? (yes/no) ",
    func: Callable[[Any], Any] = sys.exit,
    args: Iterable = None,
    kwargs: dict = None,
):
    # pylint: disable=comparison-with-callable
    """
    Prompt user for confirmation before continuing program execution.

    :param message:
        An optional message to print before showing the prompt.
    :param prompt:
        The prompt to display.
    :param func:
        The function that is called when the user answers NO
    :param args:
        Positional arguments passed to the above function.
    :param kwargs:
        Keyword arguments passed to the above function.
    """
    # Check if the force env var is set
    if getenv(settings.VNET_FORCE_ENV_VAR) == "true":
        logger.debug("Yes argument passed, skipping user confirmation")
        return

    if args is None and func == sys.exit:
        args = [1]
    elif args is None:
        args = []
    if kwargs is None:
        kwargs = {}

    if message is not None:
        logger.warning(message)

    response = None
    while True:
        if response in ("y", "yes"):
            return
        if response in ("n", "no"):
            func(*args, **kwargs)
            return
        if response is not None:
            print("Please answer yes or no.")
        response = input(prompt).lower()


def generate_bash_completion_script() -> str:
    """
    Generates the contents of the bash completion script that can be used with VNet-manager
    :returns: str: The bash completion script that can be written to disk
    """
    logger.info("Generating bash completion script")
    template = settings.VNET_BASH_COMPLETION_TEMPLATE
    return template.format(options=" ".join(settings.get("VALID_ACTIONS", "")), name=settings.get("PYTHON_PACKAGE_NAME", "vnet-manager"))



from random import randint


def random_mac_generator() -> str:
    """
    Generates a random MAC address
    :return: str: The generated MAC address
    """
    # From: https://stackoverflow.com/a/43546406/8632038
    # Put 0x02 at the start to make it a locally administered address
    return "02:00:00:%02x:%02x:%02x" % (randint(0, 255), randint(0, 255), randint(0, 255))



from unittest.mock import Mock
from os import environ, EX_NOPERM
from logging import INFO, DEBUG

from vnet_manager.tests import VNetTestCase
from vnet_manager.conf import settings
from vnet_manager.vnet_manager import main

default_args = ["list", "config"]


class TestVNetManagerMain(VNetTestCase):
    def setUp(self) -> None:
        self.setup_console_logging = self.set_up_patch("vnet_manager.vnet_manager.setup_console_logging")
        self.check_for_root_user = self.set_up_patch("vnet_manager.vnet_manager.check_for_root_user")
        self.manager = Mock()
        self.action_manager = self.set_up_patch("vnet_manager.vnet_manager.ActionManager")
        self.action_manager.return_value = self.manager
        self.environ = dict(environ)

    def tearDown(self) -> None:
        environ.clear()
        environ.update(self.environ)

    def test_main_sets_vnet_force_env_variable_to_false_by_default(self):
        main(default_args)
        self.assertEqual(environ[settings.VNET_FORCE_ENV_VAR], "false")

    def test_main_sets_vnet_force_env_variable_to_true_when_yes_passed(self):
        main(default_args + ["--yes"])
        self.assertEqual(environ[settings.VNET_FORCE_ENV_VAR], "true")

    def test_main_calls_setup_console_logging(self):
        main(default_args)
        self.setup_console_logging.assert_called_once_with(verbosity=INFO)

    def test_main_call_setup_console_logging_with_verbose(self):
        main(default_args + ["--verbose"])
        self.setup_console_logging.assert_called_once_with(verbosity=DEBUG)

    def test_main_calls_check_for_root_user(self):
        main(default_args)
        self.check_for_root_user.assert_called_once_with()

    def test_main_returns_non_zero_exit_code_when_no_root_user_detected(self):
        self.check_for_root_user.return_value = False
        ret = main(default_args)
        self.assertEqual(ret, EX_NOPERM)
        self.assertFalse(self.action_manager.called)

    def test_main_calls_action_manager(self):
        main(default_args)
        self.action_manager.assert_called_once_with(base_image=False, config_path="config", no_hosts=False, sniffer=False)

    def test_main_calls_action_manager_with_base_image(self):
        main(["destroy", "config", "--base-image"])
        self.action_manager.assert_called_once_with(base_image=True, config_path="config", no_hosts=False, sniffer=False)

    def test_main_calls_action_manager_with_no_hosts(self):
        main(["create", "config", "--no-hosts"])
        self.action_manager.assert_called_once_with(base_image=False, config_path="config", no_hosts=True, sniffer=False)

    def test_main_calls_action_manager_with_sniffer(self):
        main(["start", "config", "--sniffer"])
        self.action_manager.assert_called_once_with(base_image=False, config_path="config", no_hosts=False, sniffer=True)

    def test_main_sets_manager_machine_attribute(self):
        main(default_args + ["--machines", "test1", "test2"])
        self.assertEqual(self.manager.machines, ["test1", "test2"])

    def test_main_returns_action_manager_execute_return_value(self):
        self.manager.execute.return_value = 42
        ret = main(default_args)
        self.assertEqual(ret, 42)



import logging
import logging.config
from argparse import Namespace

from vnet_manager.conf import settings


def get_logging_verbosity(args: Namespace, default_verbosity: int = settings.LOGGING_DEFAULT_VERBOSITY) -> int:
    """
    Get the logging verbosity based upon any passed verbosity arguments
    :param args: Namespace, the parsed command line arguments
    :param default_verbosity: int, the initial verbosity to manipulate
    :return: int: The logging verbosity setting
    """
    logging_verbs = {0: logging.CRITICAL, 1: logging.ERROR, 2: logging.WARNING, 3: logging.INFO, 4: logging.DEBUG}
    # Compare the amount of verbosity args against the default verbosity
    print(default_verbosity, args)
    mod = default_verbosity + args.verbose - args.quite
    # Return the appropriate logging level (must be within the defined logging levels)
    return logging_verbs[max(min(len(logging_verbs) - 1, mod), 0)]


def setup_console_logging(verbosity: int = logging.INFO) -> None:
    """
    :param int verbosity: Verbosity level logging.<verbosity>
    """
    settings.LOGGING["handlers"]["console"]["level"] = verbosity
    settings.LOGGING["handlers"]["syslog"]["level"] = verbosity
    logging.config.dictConfig(settings.LOGGING)



from argparse import Namespace, ArgumentParser
from typing import Sequence

from vnet_manager.conf import settings


def parse_vnet_args(args: Sequence = None) -> Namespace:
    parser = ArgumentParser(description="VNet-manager a virtual network manager - manages containers to create virtual networks")
    parser.add_argument(
        "action",
        choices=sorted(settings.VALID_ACTIONS),
        help="The action to preform on the virtual network, use '<action> help' for information about that action",
    )
    parser.add_argument("config", help="The yaml config file to use", nargs="?", default="default")

    # Options
    parser.add_argument(
        "-m",
        "--machines",
        nargs="*",
        help="Just apply the actions on the following machine names " "(default is all machines defined in the config file)",
    )
    parser.add_argument("-y", "--yes", action="store_true", help="Answer yes to all questions")
    parser.add_argument("-nh", "--no-hosts", action="store_true", help="Disable creation of /etc/hosts")

    start_group = parser.add_argument_group("Start options", "These options can be specified for the start action")
    start_group.add_argument("-s", "--sniffer", action="store_true", help="Start a TCPdump sniffer on the VNet interfaces")

    destroy_group = parser.add_argument_group("Destroy options", "These options can be specified for the destroy action")
    destroy_group.add_argument("-b", "--base-image", action="store_true", help="Destroy the base image instead of the machines")

    logging_group = parser.add_argument_group("Verbosity options", "Control output verbosity (can be supplied multiple times)")
    logging_group.add_argument("-v", "--verbose", action="count", default=0, help="Be more verbose")
    logging_group.add_argument("-q", "--quite", action="count", default=0, help="Be more quite")
    return validate_argument_sanity(parser.parse_args(args=args), parser)


def validate_argument_sanity(args: Namespace, parser: ArgumentParser) -> Namespace:
    """
    Validates the passed arguments for sanity
    :param args: Namespace, The already processed user arguments
    :param parser: ArgumentParser, The parser object
    :return: Namespace, The validated arguments
    :raises: SystemExit, if arguments are not sane
    """
    # User input sanity checks
    if args.action == "status":
        # For people who are used to status calls
        args.action = "show"
    if args.config == "default" and args.action in settings.CONFIG_REQUIRED_ACTIONS:
        parser.error("This action requires a config file to be passed")
    if args.sniffer and not args.action == "start":
        parser.error("The sniffer option only makes sense with the 'start' action")
    if args.base_image and not args.action == "destroy":
        parser.error("The base_image option only makes sense with the 'destroy' action")
    if args.no_hosts and not args.action == "create":
        parser.error("The no_hosts option only makes sense with the 'create' action")
    return args



from os import EX_OK, EX_USAGE
from unittest.mock import MagicMock

from vnet_manager.tests import VNetTestCase
from vnet_manager.actions.manager import ActionManager
from vnet_manager.conf import settings


class TestActionManager(VNetTestCase):
    def setUp(self) -> None:
        self.get_config = self.set_up_patch("vnet_manager.actions.manager.get_config")
        self.validator = MagicMock()
        self.validate = self.set_up_patch("vnet_manager.actions.manager.ValidateConfig")
        self.validate.return_value = self.validator
        self.validator.updated_config = {}
        self.show_version = self.set_up_patch("vnet_manager.actions.manager.show_version")
        self.show_status = self.set_up_patch("vnet_manager.actions.manager.show_status")
        self.show_status_interfaces = self.set_up_patch("vnet_manager.actions.manager.show_vnet_interface_status")
        self.show_status_interfaces_veth = self.set_up_patch("vnet_manager.actions.manager.show_vnet_veth_interface_status")
        self.bring_up_vnet_interfaces = self.set_up_patch("vnet_manager.actions.manager.bring_up_vnet_interfaces")
        self.change_machine_status = self.set_up_patch("vnet_manager.actions.manager.change_machine_status")
        self.bring_down_vnet_interfaces = self.set_up_patch("vnet_manager.actions.manager.bring_down_vnet_interfaces")
        self.ensure_vnet_lxc_environment = self.set_up_patch("vnet_manager.actions.manager.ensure_vnet_lxc_environment")
        self.create_machines = self.set_up_patch("vnet_manager.actions.manager.create_machines")
        self.put_files_on_machine = self.set_up_patch("vnet_manager.actions.manager.put_files_on_machine")
        self.generate_vnet_hosts_file = self.set_up_patch("vnet_manager.actions.manager.generate_vnet_hosts_file")
        self.place_vnet_hosts_file_on_machines = self.set_up_patch("vnet_manager.actions.manager.place_vnet_hosts_file_on_machines")
        self.enable_type_specific_machine_configuration = self.set_up_patch(
            "vnet_manager.actions.manager.enable_type_specific_machine_configuration"
        )
        self.request_confirmation = self.set_up_patch("vnet_manager.actions.manager.request_confirmation")
        self.destroy_lxc_image = self.set_up_patch("vnet_manager.actions.manager.destroy_lxc_image")
        self.destroy_machines = self.set_up_patch("vnet_manager.actions.manager.destroy_machines")
        self.delete_vnet_interfaces = self.set_up_patch("vnet_manager.actions.manager.delete_vnet_interfaces")
        self.cleanup_vnet_lxc_environment = self.set_up_patch("vnet_manager.actions.manager.cleanup_vnet_lxc_environment")
        self.display_help_for_action = self.set_up_patch("vnet_manager.actions.manager.display_help_for_action")
        self.isfile = self.set_up_patch("vnet_manager.actions.manager.isfile")
        self.isdir = self.set_up_patch("vnet_manager.actions.manager.isdir")
        self.get_yaml_file_from_disk_path = self.set_up_patch("vnet_manager.actions.manager.get_yaml_files_from_disk_path")
        self.get_yaml_file_from_disk_path.return_value = ["file1"]

    def test_action_raises_not_implemented_error_if_unsupported_action(self):
        manager = ActionManager()
        with self.assertRaises(NotImplementedError):
            manager.execute("not_working")

    def test_action_manager_calls_show_version_with_version_action(self):
        manager = ActionManager()
        ret = manager.execute("version")
        self.show_version.assert_called_once_with()
        self.assertEqual(ret, EX_OK)

    def test_action_manager_calls_display_help_for_action_when_action_has_been_called_with_help_as_second_parameter(self):
        manager = ActionManager(config_path="help")
        ret = manager.execute("start")
        self.display_help_for_action.assert_called_once_with("start")
        self.assertEqual(ret, EX_OK)

    def test_action_manager_calls_get_config(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.get_config.assert_called_once_with("blaap")

    def test_action_manager_calls_validate_config(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.validate.called_once_with()

    def test_action_manager_called_validator_function(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.validator.validate.assert_called_once()

    def test_action_manager_returns_usage_exit_code_if_validator_unsuccessful(self):
        self.validator.config_validation_successful = False
        manager = ActionManager(config_path="blaap")
        ret = manager.execute("show")
        self.assertEqual(ret, EX_USAGE)
        self.assertFalse(self.show_status.called)

    def test_action_manager_calls_show_status_with_list_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.show_status.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_show_vnet_interface_status_with_list_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.show_status_interfaces.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_does_not_call_vnet_veth_interface_status_with_list_action_if_veths_config_not_present(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.assertFalse(self.show_status_interfaces_veth.called)

    def test_action_manager_calls_show_vnet_veth_interface_status_with_list_action(self):
        self.validator.updated_config["veths"] = "jajaja"
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.show_status_interfaces_veth.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_get_yaml_file_from_disk_path_with_list_action(self):
        self.isfile.return_value = False
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.get_yaml_file_from_disk_path.assert_called_once_with("blaap")

    def test_action_manager_calls_show_status_with_return_values_of_get_yaml_files(self):
        self.get_yaml_file_from_disk_path.return_value = ["file1", "file2", "file3"]
        self.isfile.return_value = False
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.assertEqual(3, self.show_status.call_count)

    def test_action_manager_does_nothing_when_not_file_and_not_dir_with_list_action(self):
        self.isfile.return_value = False
        self.isdir.return_value = False
        manager = ActionManager(config_path="blaap")
        manager.execute("list")
        self.assertFalse(self.get_yaml_file_from_disk_path.called)
        self.assertFalse(self.show_status.called)

    def test_action_manager_calls_show_status_with_show_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("show")
        self.show_status.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_show_vnet_interface_status_with_show_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("show")
        self.show_status_interfaces.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_does_not_call_vnet_veth_interface_status_with_show_action_if_veths_config_not_present(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("show")
        self.assertFalse(self.show_status_interfaces_veth.called)

    def test_action_manager_calls_show_vnet_veth_interface_status_with_show_action(self):
        self.validator.updated_config["veths"] = "jajaja"
        manager = ActionManager(config_path="blaap")
        manager.execute("show")
        self.show_status_interfaces_veth.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_returns_ex_ok_after_action_has_been_completed(self):
        manager = ActionManager(config_path="blaap")
        ret = manager.execute("list")
        self.assertEqual(ret, EX_OK)

    def test_action_manager_calls_bring_up_vnet_interfaces_with_start_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("start")
        self.bring_up_vnet_interfaces.assert_called_once_with(self.validator.updated_config, sniffer=False)
        self.assertFalse(self.bring_down_vnet_interfaces.called)

    def test_action_manager_calls_bring_up_vnet_interfaces_with_start_action_and_sniffer(self):
        manager = ActionManager(config_path="blaap", sniffer=True)
        manager.execute("start")
        self.bring_up_vnet_interfaces.assert_called_once_with(self.validator.updated_config, sniffer=True)

    def test_action_manager_calls_change_machine_status_with_start_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("start")
        self.change_machine_status.assert_called_once_with(self.validator.updated_config, machines=None, status="start")

    def test_action_manager_calls_change_machine_status_with_start_action_and_machines(self):
        manager = ActionManager(config_path="blaap")
        manager.machines = ["machine"]
        manager.execute("start")
        self.change_machine_status.assert_called_once_with(self.validator.updated_config, machines=["machine"], status="start")

    def test_action_manager_calls_change_machine_status_with_stop_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("stop")
        self.change_machine_status.assert_called_once_with(self.validator.updated_config, machines=None, status="stop")

    def test_action_manager_calls_change_machine_status_with_stop_action_and_machines(self):
        manager = ActionManager(config_path="blaap")
        manager.machines = ["machine"]
        manager.execute("stop")
        self.change_machine_status.assert_called_once_with(self.validator.updated_config, machines=["machine"], status="stop")

    def test_action_manager_calls_bring_down_vnet_interfaces_with_stop_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("stop")
        self.bring_down_vnet_interfaces.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_does_not_call_bring_down_vnet_interfaces_with_stop_action_and_machines(self):
        manager = ActionManager(config_path="blaap")
        manager.machines = ["machine"]
        manager.execute("stop")
        self.assertFalse(self.bring_down_vnet_interfaces.called)

    def test_action_manager_calls_ensure_vnet_lxc_environment_with_create_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("create")
        self.ensure_vnet_lxc_environment.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_create_machines_with_create_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("create")
        self.create_machines.assert_called_once_with(self.validator.updated_config, machines=None)

    def test_action_manager_calls_create_machines_with_create_action_and_machines(self):
        manager = ActionManager(config_path="blaap")
        manager.machines = ["machine"]
        manager.execute("create")
        self.create_machines.assert_called_once_with(self.validator.updated_config, machines=["machine"])

    def test_action_manager_calls_put_files_on_machine_with_create_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("create")
        self.put_files_on_machine.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_place_vnet_hosts_file_on_machines_with_create_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("create")
        self.place_vnet_hosts_file_on_machines.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_generate_vnet_hosts_file_with_create_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("create")
        self.generate_vnet_hosts_file.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_enable_type_specific_machine_configuration(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("create")
        self.enable_type_specific_machine_configuration.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_ensure_vnet_lxc_environment_with_create_action_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.execute("create")
        self.ensure_vnet_lxc_environment.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_create_machines_with_create_action_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.execute("create")
        self.create_machines.assert_called_once_with(self.validator.updated_config, machines=None)

    def test_action_manager_calls_create_machines_with_create_action_and_machines_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.machines = ["machine"]
        manager.execute("create")
        self.create_machines.assert_called_once_with(self.validator.updated_config, machines=["machine"])

    def test_action_manager_calls_put_files_on_machine_with_create_action_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.execute("create")
        self.put_files_on_machine.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_place_vnet_hosts_file_on_machines_with_create_action_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.execute("create")
        self.assertFalse(self.place_vnet_hosts_file_on_machines.called)

    def test_action_manager_calls_generate_vnet_hosts_file_with_create_action_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.execute("create")
        self.assertFalse(self.generate_vnet_hosts_file.called)

    def test_action_manager_calls_enable_type_specific_machine_configuration_and_nohosts(self):
        manager = ActionManager(config_path="blaap", no_hosts=True)
        manager.execute("create")
        self.enable_type_specific_machine_configuration.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_calls_destroy_machines_with_destroy_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("destroy")
        self.destroy_machines.assert_called_once_with(self.validator.updated_config, machines=None)

    def test_action_manager_calls_cleanup_vnet_lxc_environment_with_clean_action(self):
        manager = ActionManager()
        manager.execute("clean")
        self.cleanup_vnet_lxc_environment.assert_called_once_with()

    def test_action_manager_calls_destroy_machines_with_destroy_action_and_machines(self):
        manager = ActionManager(config_path="blaap")
        manager.machines = ["machine"]
        manager.execute("destroy")
        self.destroy_machines.assert_called_once_with(self.validator.updated_config, machines=["machine"])
        self.assertFalse(self.destroy_lxc_image.called)

    def test_action_manager_calls_delete_vnet_interfaces_with_destroy_action(self):
        manager = ActionManager(config_path="blaap")
        manager.execute("destroy")
        self.delete_vnet_interfaces.assert_called_once_with(self.validator.updated_config)

    def test_action_manager_does_not_call_delete_vnet_interfaces_with_destroy_action_and_machines(self):
        manager = ActionManager(config_path="blaap")
        manager.machines = ["machine"]
        manager.execute("destroy")
        self.assertFalse(self.delete_vnet_interfaces.called)

    def test_action_manager_calls_destroy_lxc_image_with_destroy_action_and_base_image(self):
        manager = ActionManager(config_path="blaap", base_image=True)
        manager.execute("destroy")
        self.destroy_lxc_image.assert_called_once_with(settings.LXC_BASE_IMAGE_ALIAS, by_alias=True)
        self.assertFalse(self.destroy_machines.called)

    def test_action_manager_calls_request_confirmation_with_destroy_action_and_base_image(self):
        manager = ActionManager(config_path="blaap", base_image=True)
        manager.execute("destroy")
        self.request_confirmation.assert_called_once_with(prompt="Are you sure you want to delete the VNet base images (y/n)? ")

    def test_action_manager_has_machine_property(self):
        manager = ActionManager()
        manager.machines = ["1", "2", "3"]
        self.assertEqual(manager.machines, ["1", "2", "3"])
        self.assertIsInstance(ActionManager.machines, property)



from unittest.mock import Mock, call
from copy import deepcopy

from vnet_manager.tests import VNetTestCase
from vnet_manager.config.validate import ValidateConfig
from vnet_manager.conf import settings


class TestValidateConfigClass(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.switch_config = Mock()
        self.validator.validate_switch_config = self.switch_config
        self.machine_config = Mock()
        self.validator.validate_machine_config = self.machine_config
        self.veth_config = Mock()
        self.validator.validate_veth_config = self.veth_config

    def test_validate_class_returns_all_ok_on_init(self):
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_class_ran_zero_validators_on_init(self):
        self.assertEqual(self.validator.validators_ran, 0)

    def test_validate_class_returns_proper_string_message(self):
        self.assertEqual(str(self.validator), "VNet config validator, current_state: OK, amount of validators run: 0")

    def test_validate_class_returns_original_config_on_init(self):
        self.assertEqual(self.validator.updated_config, settings.CONFIG)

    def test_validate_function_calls_standard_validator_functions(self):
        self.validator.validate()
        self.switch_config.assert_called_once_with()
        self.machine_config.assert_called_once_with()

    def test_validate_function_does_not_call_veth_validator_when_not_present_in_config(self):
        del self.validator.config["veths"]
        self.validator.validate()
        self.assertFalse(self.veth_config.called)

    def test_validate_function_calls_veth_validator_when_veths_in_config(self):
        self.validator.validate()
        self.veth_config.assert_called_once_with()


class TestValidateConfigValidateSwitchConfig(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")

    def test_validate_switch_config_runs_ok_with_good_config(self):
        self.validator.validate_switch_config()
        self.assertTrue(self.validator.config_validation_successful)
        self.assertGreater(self.validator.validators_ran, 0)

    def test_validate_switch_config_fails_when_switch_config_not_present(self):
        del self.validator.config["switches"]
        self.validator.validate_switch_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with("Config item 'switches' missing{}".format(self.validator.default_message))

    def test_validate_switch_config_fails_when_switch_config_not_a_int(self):
        self.validator.config["switches"] = "os3"
        self.validator.validate_switch_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Config item 'switches: {}' does not seem to be an integer{}".format(
                self.validator.config["switches"], self.validator.default_message
            )
        )


class TestValidateConfigValidateMachineConfig(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")
        self.validate_files = Mock()
        self.validate_interfaces = Mock()
        self.validator.validate_interface_config = self.validate_interfaces
        self.validator.validate_machine_files_parameters = self.validate_files
        self.validate_vlan_config = self.set_up_patch("vnet_manager.config.validate.ValidateConfig.validate_vlan_config")
        self.validate_bridge_config = self.set_up_patch("vnet_manager.config.validate.ValidateConfig.validate_machine_bridge_config")

    def test_validate_machine_config_runs_ok_with_good_config(self):
        self.validator.validate_machine_config()
        self.assertTrue(self.validator.config_validation_successful)
        self.assertGreater(self.validator.validators_ran, 0)

    def test_validate_machine_config_fails_when_machine_config_not_present(self):
        del self.validator.config["machines"]
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with("Config item 'machines' missing{}".format(self.validator.default_message))

    def test_validate_machine_config_fails_when_machine_config_not_a_dict(self):
        self.validator.config["machines"] = 42
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Machines config is not a dict, this means the user config is incorrect{}".format(self.validator.default_message)
        )

    def test_validate_machine_config_fails_when_machine_type_not_present(self):
        del self.validator.config["machines"]["router100"]["type"]
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with("Type not found for machine router100{}".format(self.validator.default_message))

    def test_validate_machine_config_fails_when_machine_type_not_in_supported_machine_types(self):
        self.validator.config["machines"]["router100"]["type"] = "banana"
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Type banana for machine router100 unsupported. I only support the following types: {}{}".format(
                settings.SUPPORTED_MACHINE_TYPES, self.validator.default_message
            )
        )

    def test_validate_machine_config_fails_when_machine_files_not_a_dict(self):
        self.validator.config["machines"]["router100"]["files"] = "banana"
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Files directive for machine router100 is not a dict{}".format(self.validator.default_message)
        )

    def test_validate_machine_config_succeeds_when_machine_files_not_present(self):
        del self.validator.config["machines"]["router100"]["files"]
        del self.validator.config["machines"]["router101"]["files"]
        del self.validator.config["machines"]["host102"]["files"]
        self.validator.validate_machine_config()
        self.assertTrue(self.validator.config_validation_successful)
        self.assertFalse(self.validate_files.called)

    def test_validate_machine_config_calls_validate_files(self):
        self.validator.validate_machine_config()
        calls = [call(machine) for machine in self.validator.config["machines"].keys()]
        self.validate_files.assert_has_calls(calls)

    def test_validate_machine_config_fails_if_interfaces_not_in_machine_config(self):
        del self.validator.config["machines"]["router100"]["interfaces"]
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Machine router100 does not appear to have any interfaces{}".format(self.validator.default_message)
        )

    def test_validate_machine_config_fails_if_interfaces_is_not_a_dict(self):
        self.validator.config["machines"]["router100"]["interfaces"] = 42
        self.validator.config["machines"]["router101"]["interfaces"] = 42
        self.validator.config["machines"]["host102"]["interfaces"] = 42
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        calls = [
            call(
                "The interfaces for machine {} are not given as a dict, this usually means a typo in the config{}".format(
                    machine, self.validator.default_message
                )
            )
            for machine in self.validator.config["machines"].keys()
        ]
        self.logger.error.assert_has_calls(calls)
        self.assertFalse(self.validate_interfaces.called)

    def test_validate_machine_config_calls_validate_interface_config(self):
        self.validator.validate_machine_config()
        calls = [call(machine) for machine in self.validator.config["machines"].keys()]
        self.validate_interfaces.assert_has_calls(calls)

    def test_validate_machine_config_calls_validate_vlan_config_for_each_vlan_interface(self):
        # We have only one VLAN interface defined in the settings
        self.validator.validate_machine_config()
        self.validate_vlan_config.assert_called_once_with("router100")

    def test_validate_machine_config_does_not_call_validate_vlan_config_if_not_vlans_present(self):
        del self.validator.config["machines"]["router100"]["vlans"]
        self.validator.validate_machine_config()
        self.assertFalse(self.validate_vlan_config.called)

    def test_validate_machine_config_fails_if_vlans_is_not_a_dict(self):
        self.validator.config["machines"]["router100"]["vlans"] = 1337
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Machine router100 has a VLAN config but it does not appear to be a dict, "
            "this usually means a typo in the config{}".format(self.validator.default_message)
        )

    def test_validate_machine_config_does_not_call_validate_bridge_config_if_no_bridges(self):
        del self.validator.config["machines"]["router100"]["bridges"]
        self.validator.validate_machine_config()
        self.assertFalse(self.validate_bridge_config.called)

    def test_validate_machine_config_calls_validate_bridge_config_for_machines_with_bridges(self):
        self.validator.validate_machine_config()
        self.validate_bridge_config.assert_called_once_with("router100")

    def test_validate_machine_config_fails_if_bridge_config_is_not_a_dict(self):
        self.validator.config["machines"]["router100"]["bridges"] = 1337
        self.validator.validate_machine_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once()


class TestValidateConfigValidateMachineFilesParameters(VNetTestCase):
    def setUp(self) -> None:
        # Use VALIDATED_CONFIG so we have the config_dir
        self.validator = ValidateConfig(deepcopy(settings.VALIDATED_CONFIG))
        self.is_dir = self.set_up_patch("vnet_manager.config.validate.isdir")
        self.is_dir.return_value = False
        self.is_file = self.set_up_patch("vnet_manager.config.validate.isfile")
        self.is_file.return_value = False
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")

    def test_validate_machine_file_parameters_fails_when_no_file_or_dir_found(self):
        self.validator.validate_machine_files_parameters("router100")
        self.assertFalse(self.validator.config_validation_successful)
        self.assertTrue(self.logger.error.called)

    def test_validate_machine_file_parameters_is_ok_when_is_dir_is_true(self):
        self.is_dir.return_value = True
        self.validator.validate_machine_files_parameters("router100")
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_machine_file_parameters_is_ok_when_is_file_is_true(self):
        self.is_file.return_value = True
        self.validator.validate_machine_files_parameters("router100")
        self.assertTrue(self.validator.config_validation_successful)


class TestValidateConfigValidateInterfaceConfig(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")

    def test_validate_interface_config_runs_ok_with_good_config(self):
        self.validator.validate_interface_config("router100")
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_interface_config_runs_ok_if_ipv4_not_present(self):
        del self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["ipv4"]
        self.validator.validate_interface_config("router100")
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_interface_config_fails_when_ipv4_not_valid(self):
        self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["ipv4"] = "255.255.256.257"
        self.validator.validate_interface_config("router100")
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once()

    def test_validate_interface_config_does_not_fail_when_ipv6_not_present(self):
        del self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["ipv6"]
        self.validator.validate_interface_config("router100")
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_interface_config_fails_when_ipv6_not_valid(self):
        self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["ipv6"] = "2001:h80:1:2d96::f1a5/64"
        self.validator.validate_interface_config("router100")
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once()

    def test_validate_interface_config_fails_when_bridge_not_present(self):
        del self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["bridge"]
        self.validator.validate_interface_config("router100")
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "bridge keyword missing on interface eth12 for machine router100{}".format(self.validator.default_message)
        )

    def test_validate_interface_config_fails_when_bridge_not_a_int(self):
        self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["bridge"] = "42"
        self.validator.validate_interface_config("router100")
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Invalid bridge number detected for interface eth12 on machine router100. "
            "The bridge keyword should correspond to the interface number of the vnet bridge to connect to "
            "(starting at iface number 0)"
        )

    def test_validate_interface_config_fails_when_bridge_number_higher_then_the_amount_of_switches(self):
        self.validator.config["machines"]["router100"]["interfaces"]["eth12"]["bridge"] = 3
        self.validator.validate_interface_config("router100")
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Invalid bridge number detected for interface eth12 on machine router100. "
            "The bridge keyword should correspond to the interface number of the vnet bridge to connect to "
            "(starting at iface number 0)"
        )


class TestValidateConfigValidateVethConfig(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")

    def test_validate_veth_config_does_nothing_when_veth_config_not_present(self):
        del self.validator.config["veths"]
        self.validator.validate_veth_config()
        self.assertTrue(self.validator.config_validation_successful)
        self.logger.warning.assert_called_once_with("Tried to validate veth config, but no veth config present, skipping...")

    def test_validate_veth_config_fails_when_veth_config_is_not_a_dict(self):
        self.validator.config["veths"] = 42
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Config item: 'veths' does not seem to be a dict {}".format(self.validator.default_message)
        )

    def test_validate_veth_config_fails_when_veth_config_name_if_not_a_string(self):
        self.validator.config["veths"][42] = self.validator.config["veths"].pop("vnet-veth1")
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "veth interface name: 42 does not seem to be a string{}".format(self.validator.default_message)
        )

    def test_validate_veth_config_fails_when_veth_config_values_if_not_a_dict(self):
        self.validator.config["veths"]["vnet-veth1"] = "blaap"
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "veth interface vnet-veth1 data does not seem to be a dict{}".format(self.validator.default_message)
        )

    def test_validate_veth_config_fails_when_veth_config_parameter_bridge_missing(self):
        del self.validator.config["veths"]["vnet-veth1"]["bridge"]
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "veth interface vnet-veth1 is missing the bridge parameter{}".format(self.validator.default_message)
        )

    def test_validate_veth_config_fails_when_veth_config_parameter_bridge_is_not_a_string(self):
        self.validator.config["veths"]["vnet-veth1"]["bridge"] = 42
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "veth interface vnet-veth1 bridge parameter does not seem to be a str{}".format(self.validator.default_message)
        )

    def test_validate_veth_config_fails_when_veth_config_parameter_peer_is_not_a_string(self):
        self.validator.config["veths"]["vnet-veth1"]["peer"] = 42
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "veth interface vnet-veth1 peer parameter does not seem to be a string{}".format(self.validator.default_message)
        )

    def test_validate_veth_config_fails_when_veth_config_parameter_stp_is_not_a_bool(self):
        self.validator.config["veths"]["vnet-veth1"]["stp"] = "42"
        self.validator.validate_veth_config()
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "veth interface vnet-veth1 stp parameter does not seem to be a boolean{}".format(self.validator.default_message)
        )


class TestValidateConfigValidateVLANConfig(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")
        self.machine = "router100"

    def test_validate_vlan_config_is_successful_is_everything_okay(self):
        self.validator.validate_vlan_config(self.machine)
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_vlan_config_fails_is_id_is_not_present(self):
        del self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["id"]
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "VLAN vlan.100 on machine {} is missing it's vlan id{}".format(self.machine, self.validator.default_message)
        )

    def test_validate_vlan_config_fails_if_id_is_not_castable_to_int(self):
        self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["id"] = "banaan"
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Unable to cast VLAN vlan.100 with ID banaan from machine {} to a integer{}".format(
                self.machine, self.validator.default_message
            )
        )

    def test_validate_vlan_config_fails_if_link_is_not_present(self):
        del self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["link"]
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "VLAN vlan.100 on machine {} is missing it's link attribute{}".format(self.machine, self.validator.default_message)
        )

    def test_validate_vlan_config_fails_if_link_is_not_a_string(self):
        self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["link"] = 42
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Link 42 for VLAN vlan.100 on machine {}, does not seem to be a string{}".format(self.machine, self.validator.default_message)
        )

    def test_validate_vlan_config_fails_if_link_is_not_found_in_machine_interfaces(self):
        self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["link"] = "eth1337"
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Link eth1337 for VLAN vlan.100 on machine {} does not correspond to any interfaces on "
            "the same machine{}".format(self.machine, self.validator.default_message)
        )

    def test_validate_vlan_config_does_not_check_link_in_interfaces_if_config_validation_already_failed(self):
        self.validator._all_ok = False
        self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["link"] = "eth1337"
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.logger.error.called)

    def test_validate_vlan_config_does_not_fail_if_addresses_not_in_values(self):
        del self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["addresses"]
        self.validator.validate_vlan_config(self.machine)
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_vlan_config_fails_if_invalid_address_in_addresses(self):
        self.validator.config["machines"][self.machine]["vlans"]["vlan.100"]["addresses"].append("banaan")
        self.validator.validate_vlan_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.assertTrue(
            self.logger.error.call_args_list[0].startswith("Address banaan for VLAN vlan.100 on machine {}".format(self.machine))
        )


class TestValidateConfigValidateMachineBridgeConfig(VNetTestCase):
    def setUp(self) -> None:
        self.validator = ValidateConfig(deepcopy(settings.CONFIG))
        self.logger = self.set_up_patch("vnet_manager.config.validate.logger")
        self.machine = "router100"

    def test_validate_machine_bridge_config_is_successful_with_correct_bridge_config(self):
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_machine_bridge_config_accepts_missing_ipv4(self):
        del self.validator.config["machines"][self.machine]["bridges"]["br1"]["ipv4"]
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_machine_bridge_config_accepts_missing_ipv6(self):
        del self.validator.config["machines"][self.machine]["bridges"]["br1"]["ipv6"]
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertTrue(self.validator.config_validation_successful)

    def test_validate_machine_bridge_config_fails_if_incorrect_ipv4(self):
        self.validator.config["machines"][self.machine]["bridges"]["br1"]["ipv4"] = "blaap"
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once()

    def test_validate_machine_bridge_config_fails_if_incorrect_ipv6(self):
        self.validator.config["machines"][self.machine]["bridges"]["br1"]["ipv6"] = "blaap"
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once()

    def test_validate_machine_bridge_config_fails_if_slaves_not_in_bridge_params(self):
        del self.validator.config["machines"][self.machine]["bridges"]["br1"]["slaves"]
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with("Bridge {} on machine {} does not have any slaves".format("br1", self.machine))

    def test_validate_machine_bridge_config_fails_if_slaves_param_is_not_a_list(self):
        self.validator.config["machines"][self.machine]["bridges"]["br1"]["slaves"] = "blaap"
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Slaves on bridge {} for machine {}, is not formatted as a list".format("br1", self.machine)
        )

    def test_validate_machine_bridge_config_fails_if_slave_not_present_in_interfaces_config(self):
        iface = "blaap1"
        self.validator.config["machines"][self.machine]["bridges"]["br1"]["slaves"].append(iface)
        self.validator.validate_machine_bridge_config(self.machine)
        self.assertFalse(self.validator.config_validation_successful)
        self.logger.error.assert_called_once_with(
            "Undefined slave interface {} assigned to bridge {} on machine {}".format(iface, "br1", self.machine)
        )



from unittest.mock import call
from os.path import realpath

from vnet_manager.conf import settings
from vnet_manager.tests import VNetTestCase
from vnet_manager.config.config import get_config


class TestGetConfig(VNetTestCase):
    def setUp(self) -> None:
        self.get_yaml_content = self.set_up_patch("vnet_manager.config.config.get_yaml_content")
        self.get_yaml_content.side_effect = [settings.CONFIG, settings.VALIDATED_CONFIG]
        self.config_dir = self.set_up_patch("vnet_manager.config.config.dirname")
        self.config_dir.return_value = "config_dir"
        self.maxDiff = None

    def test_get_config_calls_get_yaml_contents_with_correct_path(self):
        get_config("blaap")
        self.get_yaml_content.assert_called_once_with("blaap")

    def test_get_config_user_config_takes_precedence_over_defaults_config(self):
        config = get_config("blaap")
        del config["config_dir"]
        self.assertEqual(config, settings.CONFIG)

    def test_get_config_calls_dirname_function(self):
        get_config("blaap")
        self.config_dir.assert_called_once_with(realpath("blaap"))

    def test_get_config_sets_config_dir_based_on_dirname_function(self):
        config = get_config("blaap")
        self.assertEqual(config["config_dir"], self.config_dir.return_value)



from pylxd import client


def get_lxd_client(**kwargs) -> client.Client:
    """
    Get an LXC client.Client() with the passed parameters
    :return: pylxd.client.Client()
    """
    return client.Client(**kwargs)



from distro import codename
from logging import getLogger
from sys import modules

from vnet_manager.conf import settings


logger = getLogger(__name__)

try:
    from apt import Cache
except (ImportError, ModuleNotFoundError):
    logger.error("Python apt module not installed, please 'apt-get install python3-apt' first")
    logger.warning("Will not preform host checks without apt module")


def check_for_supported_os(provider: str) -> bool:
    """
    Checks if VNet is running on a supported OS for a provider
    :param str provider: The provider to check OS support for
    :return: bool: True if the current OS is supported, False otherwise
    """
    logger.debug("Checking if your os is supported for provider {}".format(provider))
    return codename().lower() in settings["PROVIDERS"][provider]["supported_operating_systems"]


def check_for_installed_packages(provider: str) -> bool:
    """
    Checks if the required host packages have been installed
    :param str provider: The provider to check the required host packages for
    :return bool: True if all required packages have been installed, False otherwise
    """
    logger.debug("Checking if all required host packages have been install for provider {}".format(provider))
    all_installed = True
    cache = Cache()
    if "apt" not in modules:
        logger.warning("Skipping host package rquirements check, apt module missing")
    else:
        for package in settings["PROVIDERS"][provider]["required_host_packages"]:
            if not cache[package].is_installed:
                logger.error("Required host package {} for provider {} is not installed".format(package, provider))
                all_installed = False
    return all_installed



from os.path import isfile, join
from os import walk
from yaml import safe_load
from logging import getLogger
from typing import AnyStr, List

logger = getLogger(__name__)


def get_yaml_content(path: str) -> dict:
    """
    Loads a YAML config file and returns the contents
    :param str path: The YAML file to load
    :return: dict: The YAML file contents
    """
    if not isfile(path):
        logger.error("File {} does not exist".format(path))
        raise IOError("File {} does not exist".format(path))

    logger.debug("Loading YAML values from {}".format(path))
    with open(path, "r") as fh:
        content = safe_load(fh)
    return content


def write_file_to_disk(path: str, content: AnyStr):
    """
    Write a file to disk
    Overwrites if file already exists
    :param: str: path: The filepath to write the file to
    :param: content: The content to write to the file
    """
    logger.debug("Writing content to {}".format(path))
    with open(path, "w") as fh:
        fh.write(content)


def get_yaml_files_from_disk_path(path: str, excludes_files: List[str] = None) -> List[str]:
    """
    Returns a list of yaml files from a path (recursive search)
    :param path: str: The path to search in
    :param excludes_files: list: Of filenames to exclude
    :return: list of paths: the found yaml files
    """

    def should_be_excluded(root_path, file):
        if excludes_files:
            return file in excludes_files or join(root_path, file) in excludes_files
        return False

    yaml_files = []
    logger.debug("Retrieving yaml files from path: {}".format(path))
    for root, _, files in walk(path):
        for f in files:
            if (f.endswith("yaml") or f.endswith("yml")) and not should_be_excluded(root, f):
                yaml_files.append(join(root, f))
    return yaml_files



import shlex
from os.path import join
from subprocess import DEVNULL, CalledProcessError
from unittest.mock import Mock, MagicMock, ANY, call
from copy import deepcopy

from vnet_manager.tests import VNetTestCase
from vnet_manager.operations.interface import (
    get_vnet_interface_names_from_config,
    get_machines_by_vnet_interface_name,
    show_vnet_interface_status,
    show_vnet_veth_interface_status,
    check_if_interface_exists,
    create_vnet_interface,
    create_veth_interface,
    create_vnet_interface_iptables_rules,
    configure_vnet_interface,
    configure_veth_interface,
    bring_up_vnet_interfaces,
    ensure_vnet_veth_interfaces,
    check_if_sniffer_exists,
    bring_down_vnet_interfaces,
    delete_vnet_interfaces,
    start_tcpdump_on_vnet_interface,
)
from vnet_manager.conf import settings


class TestGetVNetInterfaceNamesFromConfig(VNetTestCase):
    def test_get_vnet_interface_names_from_config_returns_a_list(self):
        self.assertIsInstance(get_vnet_interface_names_from_config(settings.CONFIG), list)

    def test_get_vnet_interface_names_from_config_is_equal_to_the_length_of_defined_switches(self):
        self.assertEqual(len(get_vnet_interface_names_from_config(settings.CONFIG)), settings.CONFIG["switches"])

    def test_get_vnet_interface_names_from_config_startswith_vnet_bridge_name(self):
        for interface in get_vnet_interface_names_from_config(settings.CONFIG):
            self.assertRegex(interface, r"^{}".format(settings.VNET_BRIDGE_NAME))


class TestGetMachinesByVNetInterfaceName(VNetTestCase):
    def test_get_machines_by_vnet_interface_name_returns_a_list(self):
        self.assertIsInstance(get_vnet_interface_names_from_config(settings.CONFIG), list)

    def test_get_machines_by_vnet_interface_name_returns_the_correct_machines_per_interface(self):
        interface_mapping = {
            settings.VNET_BRIDGE_NAME + "0": ["router100", "router101"],
            settings.VNET_BRIDGE_NAME + "1": ["router101", "host102"],
        }
        for interface in interface_mapping:
            self.assertEqual(get_machines_by_vnet_interface_name(settings.CONFIG, interface), interface_mapping[interface])


class TestShowVNetInterfaceStatus(VNetTestCase):
    def setUp(self) -> None:
        self.iproute_obj = Mock()
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute.return_value = self.iproute_obj
        self.iproute_obj.link_lookup.return_value = "dev1"
        self.ndb_obj = MagicMock()
        self.ndb = self.set_up_patch("vnet_manager.operations.interface.NDB", themock=MagicMock())
        self.check_if_sniffer_exists = self.set_up_patch("vnet_manager.operations.interface.check_if_sniffer_exists")
        self.check_if_sniffer_exists.return_value = True
        self.tabulate = self.set_up_patch("vnet_manager.operations.interface.tabulate")
        self.interfaces = self.set_up_patch("vnet_manager.operations.interface.get_vnet_interface_names_from_config")
        self.interfaces.return_value = ["vnet-br0"]

    def test_show_vnet_interface_status_calls_iproute(self):
        show_vnet_interface_status(settings.CONFIG)
        self.iproute.assert_called_once_with()

    def test_show_vnet_interface_status_calls_ndb(self):
        show_vnet_interface_status(settings.CONFIG)
        self.ndb.assert_called_once_with(log=False)

    def test_show_vnet_interfaces_status_calls_get_vnet_interface_names_from_config(self):
        show_vnet_interface_status(settings.CONFIG)
        self.interfaces.assert_called_once_with(settings.CONFIG)

    def test_show_vnet_interface_status_calls_get_machines_by_vnet_interface_name(self):
        machines = self.set_up_patch("vnet_manager.operations.interface.get_machines_by_vnet_interface_name")
        machines.return_value = []
        show_vnet_interface_status(settings.CONFIG)
        machines.assert_called_once_with(settings.CONFIG, self.interfaces.return_value[0])

    def test_show_vnet_interface_status_calls_link_lookup(self):
        show_vnet_interface_status(settings.CONFIG)
        self.iproute_obj.link_lookup.assert_called_once_with(ifname=self.interfaces.return_value[0])

    def test_show_vnet_interface_status_calls_check_if_sniffer_exists(self):
        show_vnet_interface_status(settings.CONFIG)
        self.check_if_sniffer_exists.assert_called_once_with(self.interfaces.return_value[0])

    def test_show_vnet_interface_status_calls_tabulate(self):
        show_vnet_interface_status(settings.CONFIG)
        self.tabulate.assert_called_once_with(
            [["vnet-br0", ANY, ANY, self.check_if_sniffer_exists.return_value, True, "router100, router101"]],
            headers=["Name", "Status", "L2_addr", "Sniffer", "STP", "Used by"],
            tablefmt="pretty",
        )

    def test_show_vnet_interface_status_makes_correct_output_if_interface_does_not_exist(self):
        self.iproute_obj.link_lookup.return_value = []
        show_vnet_interface_status(settings.CONFIG)
        self.assertFalse(self.check_if_sniffer_exists.called)
        self.tabulate.assert_called_once_with(
            [["vnet-br0", "NA", "NA", "NA", "NA", "router100, router101"]],
            headers=["Name", "Status", "L2_addr", "Sniffer", "STP", "Used by"],
            tablefmt="pretty",
        )

    def test_show_vnet_interface_status_displays_result_if_check_if_sniffer_exists(self):
        self.check_if_sniffer_exists.return_value = False
        show_vnet_interface_status(settings.CONFIG)
        self.tabulate.assert_called_once_with(
            [["vnet-br0", ANY, ANY, self.check_if_sniffer_exists.return_value, True, "router100, router101"]],
            headers=["Name", "Status", "L2_addr", "Sniffer", "STP", "Used by"],
            tablefmt="pretty",
        )


class TestShowVNetVethInterfaceStatus(VNetTestCase):
    def setUp(self) -> None:
        self.iproute_obj = Mock()
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute.return_value = self.iproute_obj
        self.iproute_obj.link.return_value = [
            {
                "state": "up",
                "attrs": [
                    ("IFLA_ADDRESS", "mac"),
                    ("IFLA_LINK", "dev2"),
                    ("IFLA_IFNAME", "peer"),
                    ("IFLA_MASTER", "eth0"),
                    ("IFLA_IFNAME", "master"),
                ],
            }
        ]
        self.iproute_obj.link_lookup.return_value = ["dev1"]
        self.tabulate = self.set_up_patch("vnet_manager.operations.interface.tabulate")

    def test_show_vnet_veth_interface_status_calls_iproute(self):
        show_vnet_veth_interface_status(settings.CONFIG)
        self.iproute.assert_called_once_with()

    def test_show_vnet_veth_interface_status_calls_ip_lookup(self):
        calls = [call(ifname=i) for i in settings.CONFIG["veths"]]
        show_vnet_veth_interface_status(settings.CONFIG)
        self.iproute_obj.link_lookup.assert_has_calls(calls)

    def test_show_vnet_veth_interface_status_calls_ip_link(self):
        show_vnet_veth_interface_status(settings.CONFIG)
        calls = [call("get", index="dev1"), call("get", index="dev2"), call("get", index="eth0")]
        self.iproute_obj.link.assert_has_calls(calls)

    def test_show_vnet_veth_interface_status_calls_tabulate(self):
        show_vnet_veth_interface_status(settings.CONFIG)
        self.tabulate.assert_called_once_with(
            [["vnet-veth1", "up", "mac", "peer", "peer"], ["vnet-veth0", "up", "mac", "peer", "peer"]],
            headers=["Name", "Status", "L2_addr", "Peer", "Master"],
            tablefmt="pretty",
        )

    def test_show_vnet_veth_interface_status_calls_tabulate_when_dev_does_not_exist(self):
        self.iproute_obj.link_lookup.return_value = []
        show_vnet_veth_interface_status(settings.CONFIG)
        self.tabulate.assert_called_once_with(
            [["vnet-veth1", "NA", "NA", "NA", "vnet-br1"], ["vnet-veth0", "NA", "NA", "NA", "vnet-br0"]],
            headers=["Name", "Status", "L2_addr", "Peer", "Master"],
            tablefmt="pretty",
        )


class TestCheckIfInterfaceExists(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute.return_value.link_lookup.return_value = [1]

    def test_check_if_interface_exists_calls_iproute_lookup(self):
        check_if_interface_exists("dev1")
        self.iproute.return_value.link_lookup.assert_called_once_with(ifname="dev1")

    def test_check_if_interface_exists_returns_true_if_it_exists(self):
        self.assertTrue(check_if_interface_exists("dev1"))

    def test_check_if_interface_exists_returns_false_if_it_does_not_exist(self):
        self.iproute.return_value.link_lookup.return_value = []
        self.assertFalse(check_if_interface_exists("dev1"))


class TestCreateVNetInterface(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.configure_int = self.set_up_patch("vnet_manager.operations.interface.configure_vnet_interface")

    def test_create_vnet_interface_calls_iproute_link_add(self):
        create_vnet_interface("dev1")
        self.iproute.return_value.link.assert_called_once_with("add", ifname="dev1", kind="bridge")

    def test_create_vnet_interface_calls_configure_vnet_interface(self):
        create_vnet_interface("dev1")
        self.configure_int.assert_called_once_with("dev1")


class TestCreateVethInterface(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")

    def test_create_veth_interface_calls_iproute_link_add(self):
        create_veth_interface("vnet-veth0", settings.CONFIG["veths"]["vnet-veth0"])
        self.iproute.return_value.link.assert_called_once_with("add", ifname="vnet-veth0", kind="veth", peer="vnet-veth1")

    def test_create_veth_interface_does_nothing_when_called_with_interface_without_a_peer(self):
        create_veth_interface("vneth-veth0", settings.CONFIG["veths"]["vnet-veth1"])
        self.assertFalse(self.iproute.return_value.link.called)


class TestCreateVNetInterfaceIPtablesDropRules(VNetTestCase):
    def setUp(self) -> None:
        self.check_call = self.set_up_patch("vnet_manager.operations.interface.check_call")
        self.check_call.side_effect = [CalledProcessError(1, "test"), 0]
        self.logger = self.set_up_patch("vnet_manager.operations.interface.logger")

    def test_create_vnet_interface_iptables_drop_rules_makes_correct_check_call_calls(self):
        calls = [
            call(shlex.split("iptables -C OUTPUT -o dev1 -j DROP"), stderr=DEVNULL),
            call(shlex.split("iptables -A OUTPUT -o dev1 -j DROP")),
        ]
        create_vnet_interface_iptables_rules("dev1")
        self.check_call.assert_has_calls(calls)
        self.logger.info.assert_called_once_with("Creating IPtables DROP rule to the outside world for VNet interface dev1")

    def test_create_vnet_interface_iptables_drop_rules_does_not_add_rules_if_they_already_exist(self):
        self.check_call.side_effect = [0, 0]
        create_vnet_interface_iptables_rules("dev1")
        self.check_call.assert_called_once_with(shlex.split("iptables -C OUTPUT -o dev1 -j DROP"), stderr=DEVNULL)
        self.logger.debug.assert_called_once_with("IPtables DROP rule for VNet interface dev1 already exists, skipping creation")

    def test_create_vnet_interface_iptables_drop_rules_logs_error_if_both_commands_fail(self):
        self.check_call.side_effect = [CalledProcessError(1, "test"), CalledProcessError(1, "test")]
        create_vnet_interface_iptables_rules("dev1")
        self.logger.error.assert_called_once_with("Unable to create IPtables rule, got output: None")


class TestConfigureVNetInterface(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute_obj = Mock()
        self.iproute.return_value = self.iproute_obj
        self.iproute_obj.link_lookup.return_value = [1]
        self.rand_mac = self.set_up_patch("vnet_manager.operations.interface.random_mac_generator")

    def test_configure_vnet_interfaces_calls_ip_route(self):
        configure_vnet_interface("test")
        self.iproute.assert_called_once_with()

    def test_configure_vnet_interface_looks_up_passed_interface(self):
        configure_vnet_interface("test")
        self.iproute_obj.link_lookup.assert_called_once_with(ifname="test")

    def test_configure_vnet_interface_calls_random_mac_generator(self):
        configure_vnet_interface("test")
        self.rand_mac.assert_called_once_with()

    def test_configure_vnet_interface_makes_correct_ip_set_calls(self):
        calls = [
            call("set", index=1, state="down"),
            call("set", index=1, address=self.rand_mac.return_value),
            call("set", index=1, state="up"),
        ]
        configure_vnet_interface("test")
        self.iproute_obj.link.assert_has_calls(calls)


class TestConfigureVethInterface(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute_obj = Mock()
        self.iproute.return_value = self.iproute_obj
        self.iproute_obj.link_lookup.side_effect = [[1], [2]]
        self.data = settings.CONFIG["veths"]["vnet-veth1"]

    def test_configure_veth_interface_calls_ip_route(self):
        configure_veth_interface("test", self.data)
        self.iproute.assert_called_once_with()

    def test_configure_veth_interface_makes_correct_ip_lookup_calls(self):
        calls = [call(ifname="test"), call(ifname=settings.CONFIG["veths"]["vnet-veth1"]["bridge"])]
        configure_veth_interface("test", self.data)
        self.iproute_obj.link_lookup.assert_has_calls(calls)

    def test_configure_veth_interface_calls_ip_link(self):
        configure_veth_interface("test", self.data)
        self.iproute_obj.link.assert_called_once_with("set", index=1, master=2)


class TestBringUpVNetInterfaces(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute_obj = Mock()
        self.iproute.return_value = self.iproute_obj
        self.get_vnet_interface_names = self.set_up_patch("vnet_manager.operations.interface.get_vnet_interface_names_from_config")
        self.get_vnet_interface_names.return_value = ["int1", "int2"]
        self.check_if_interface_exists = self.set_up_patch("vnet_manager.operations.interface.check_if_interface_exists")
        self.check_if_interface_exists.return_value = False
        self.create_vnet_interface = self.set_up_patch("vnet_manager.operations.interface.create_vnet_interface")
        self.create_vnet_interface_block_rules = self.set_up_patch("vnet_manager.operations.interface.create_vnet_interface_iptables_rules")
        self.check_if_sniffer_exists = self.set_up_patch("vnet_manager.operations.interface.check_if_sniffer_exists")
        self.check_if_sniffer_exists.return_value = False
        self.start_tcpdump_on_interface = self.set_up_patch("vnet_manager.operations.interface.start_tcpdump_on_vnet_interface")
        self.ensure_vnet_veth_interfaces = self.set_up_patch("vnet_manager.operations.interface.ensure_vnet_veth_interfaces")
        self.config = deepcopy(settings.CONFIG)
        self.expected_vnet_interface_calls = [call(i) for i in self.get_vnet_interface_names.return_value]

    def test_bring_up_vnet_interfaces_calls_ip_route(self):
        bring_up_vnet_interfaces(self.config)
        self.iproute.assert_called_once_with()

    def test_bring_up_vnet_interfaces_calls_get_vnet_interface_names_from_config(self):
        bring_up_vnet_interfaces(self.config)
        self.get_vnet_interface_names.assert_called_once_with(self.config)

    def test_bring_up_vnet_interfaces_calls_check_if_interface_exists_with_interface_names(self):
        bring_up_vnet_interfaces(self.config)
        self.check_if_interface_exists.assert_has_calls(self.expected_vnet_interface_calls)

    def test_bring_up_vnet_interfaces_calls_create_vnet_interface_with_interface_names(self):
        bring_up_vnet_interfaces(self.config)
        self.create_vnet_interface.assert_has_calls(self.expected_vnet_interface_calls)

    def test_bring_up_vnet_interfaces_does_not_call_create_interface_if_the_interface_already_exists(self):
        self.check_if_interface_exists.return_value = True
        bring_up_vnet_interfaces(self.config)
        self.assertFalse(self.create_vnet_interface.called)

    def test_bring_up_vnet_interfaces_calls_create_vnet_interface_iptables_rules(self):
        bring_up_vnet_interfaces(self.config)
        self.create_vnet_interface_block_rules.assert_has_calls(self.expected_vnet_interface_calls)

    def test_bring_up_vnet_interfaces_calls_ip_link_to_bring_up_interfaces(self):
        calls = [call("set", ifname=i, state="up") for i in self.get_vnet_interface_names.return_value]
        bring_up_vnet_interfaces(self.config)
        self.iproute_obj.link.assert_has_calls(calls)

    def test_bring_up_vnet_interfaces_does_not_create_sniffer_by_default(self):
        bring_up_vnet_interfaces(self.config)
        self.assertFalse(self.start_tcpdump_on_interface.called)

    def test_bring_up_vnet_interfaces_calls_sniffer_when_sniffer_argument_passed(self):
        bring_up_vnet_interfaces(self.config, sniffer=True)
        self.start_tcpdump_on_interface.assert_has_calls(self.expected_vnet_interface_calls)

    def test_bring_up_vnet_interfaces_calls_check_if_sniffer_exists(self):
        bring_up_vnet_interfaces(self.config, sniffer=True)
        self.check_if_sniffer_exists.assert_has_calls(self.expected_vnet_interface_calls)

    def test_bring_up_vnet_interfaces_does_not_call_start_sniffer_when_the_sniffer_already_exists(self):
        self.check_if_sniffer_exists.return_value = True
        bring_up_vnet_interfaces(self.config, sniffer=True)
        self.assertFalse(self.start_tcpdump_on_interface.called)

    def test_bring_up_vnet_interfaces_calls_ensure_vnet_veth_interfaces(self):
        bring_up_vnet_interfaces(self.config)
        self.ensure_vnet_veth_interfaces.assert_called_once_with(self.config)

    def test_bring_up_vnet_interfaces_does_not_calls_ensure_vnet_veth_interface_if_no_veth_interfaces_present_in_config(self):
        del self.config["veths"]
        bring_up_vnet_interfaces(self.config)
        self.assertFalse(self.ensure_vnet_veth_interfaces.called)


class TestEnsureVNetVethInterfaces(VNetTestCase):
    def setUp(self) -> None:
        self.config = deepcopy(settings.CONFIG)
        self.ndb = self.set_up_patch("vnet_manager.operations.interface.NDB", themock=MagicMock())
        self.check_if_interface_exists = self.set_up_patch("vnet_manager.operations.interface.check_if_interface_exists")
        self.check_if_interface_exists.return_value = False
        self.create_veth_interface = self.set_up_patch("vnet_manager.operations.interface.create_veth_interface")
        self.configure_veth_interface = self.set_up_patch("vnet_manager.operations.interface.configure_veth_interface")
        self.configure_vnet_interface = self.set_up_patch("vnet_manager.operations.interface.configure_vnet_interface")

    def test_ensure_vnet_veth_interfaces_calls_ndb(self):
        ensure_vnet_veth_interfaces(self.config)
        self.ndb.assert_called_with(log=False)

    def test_ensure_vnet_veth_interfaces_set_stp_state_to_correct_state_according_to_config(self):
        ensure_vnet_veth_interfaces(self.config)
        calls = [call("vnet-br1"), call("vnet-br0")]
        self.assertIn(calls, self.ndb.return_value.interfaces.__getitem__.call_args_list)

    def test_ensure_vnet_veth_interfaces_does_not_call_ndb_if_stp_not_in_int_data(self):
        del self.config["veths"]["vnet-veth1"]["stp"]
        ensure_vnet_veth_interfaces(self.config)
        self.assertNotIn(call("vnet-br1"), self.ndb.return_value.interfaces.__getitem__.call_args_list)

    def test_ensure_vnet_veth_interfaces_sets_correct_stp_state_on_bridge_ints(self):
        ensure_vnet_veth_interfaces(self.config)
        calls = [call("br_stp_state", 1), call("br_stp_state", 0)]
        self.assertIn(calls, self.ndb.return_value.interfaces.__getitem__.return_value.__enter__.return_value.set.call_args_list)

    def test_ensure_vnet_veth_interfaces_checks_if_veth_interfaces_already_exist(self):
        ensure_vnet_veth_interfaces(self.config)
        calls = [call(i) for i in self.config["veths"]]
        self.check_if_interface_exists.assert_has_calls(calls)

    def test_ensure_vnet_veth_interfaces_calls_create_veth_interfaces(self):
        ensure_vnet_veth_interfaces(self.config)
        calls = [call(k, v) for k, v in self.config["veths"].items()]
        self.create_veth_interface.assert_has_calls(calls)

    def test_ensure_vnet_veth_interfaces_does_not_call_create_interfaces_if_they_already_exist(self):
        self.check_if_interface_exists.return_value = True
        self.assertFalse(self.create_veth_interface.called)

    def test_ensure_vnet_veth_interfaces_calls_configure_veth_interface(self):
        ensure_vnet_veth_interfaces(self.config)
        calls = [call(k, v) for k, v in self.config["veths"].items()]
        self.configure_veth_interface.assert_has_calls(calls)

    def test_ensure_vnet_veth_interfaces_calls_configure_vnet_interface(self):
        ensure_vnet_veth_interfaces(self.config)
        calls = [call(i) for i in self.config["veths"]]
        self.configure_vnet_interface.assert_has_calls(calls)


class TestCheckIfSnifferExists(VNetTestCase):
    def setUp(self) -> None:
        self.process = Mock()
        self.process_iter = self.set_up_patch("vnet_manager.operations.interface.process_iter")
        self.process_iter.return_value = [self.process]
        self.process.cmdline.return_value = "testprocess testargs"

    def test_check_if_sniffer_exists_returns_false_if_sniffer_does_not_exist(self):
        self.assertFalse(check_if_sniffer_exists("dev0"))

    def test_check_if_sniffer_exists_returns_true_if_sniffer_exists(self):
        self.process.cmdline.return_value = "/usr/sbin/tcpdump -i dev1 -n"
        self.assertTrue(check_if_sniffer_exists("dev1"))


class TestBringDownVNetInterfaces(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute_obj = Mock()
        self.iproute.return_value = self.iproute_obj
        self.check_if_interface_exists = self.set_up_patch("vnet_manager.operations.interface.check_if_interface_exists")
        self.config = deepcopy(settings.CONFIG)

    def test_bring_down_vnet_interfaces_calls_iproute(self):
        bring_down_vnet_interfaces(self.config)
        self.iproute.assert_called_once_with()

    def test_bring_down_vnet_interfaces_check_if_interface_exists_for_each_interface_in_config(self):
        calls = [call("vnet-veth1"), call("vnet-veth0"), call("vnet-br0"), call("vnet-br1")]
        bring_down_vnet_interfaces(self.config)
        self.check_if_interface_exists.assert_has_calls(calls)
        self.assertEqual(self.check_if_interface_exists.call_count, 4)

    def test_bring_down_vnet_interfaces_does_not_check_veth_interfaces_if_not_in_config(self):
        calls = [call("vnet-br0"), call("vnet-br1")]
        del self.config["veths"]
        bring_down_vnet_interfaces(self.config)
        self.check_if_interface_exists.assert_has_calls(calls)
        self.assertEqual(self.check_if_interface_exists.call_count, 2)

    def test_bring_down_vnet_interfaces_calls_ip_link_to_bring_down_interfaces(self):
        calls = [call("set", ifname=i, state="down") for i in ["vnet-veth1", "vnet-veth0", "vnet-br0", "vnet-br1"]]
        bring_down_vnet_interfaces(self.config)
        self.iproute_obj.link.assert_has_calls(calls)
        self.assertEqual(self.iproute_obj.link.call_count, 4)

    def test_bring_down_vnet_interfaces_down_not_bring_down_veth_interfaces_if_not_in_config(self):
        calls = [call("set", ifname=i, state="down") for i in ["vnet-br0", "vnet-br1"]]
        del self.config["veths"]
        bring_down_vnet_interfaces(self.config)
        self.iproute_obj.link.assert_has_calls(calls)
        self.assertEqual(self.iproute_obj.link.call_count, 2)

    def test_bring_down_vnet_interfaces_does_nothing_if_interfaces_do_not_exist(self):
        self.check_if_interface_exists.return_value = False
        bring_down_vnet_interfaces(self.config)
        self.assertFalse(self.iproute_obj.link.called)


class TestDeleteVNetInterfaces(VNetTestCase):
    def setUp(self) -> None:
        self.iproute = self.set_up_patch("vnet_manager.operations.interface.IPRoute")
        self.iproute_obj = Mock()
        self.iproute.return_value = self.iproute_obj
        self.check_if_interface_exists = self.set_up_patch("vnet_manager.operations.interface.check_if_interface_exists")
        self.config = deepcopy(settings.CONFIG)

    def test_delete_vnet_interfaces_calls_iproute(self):
        delete_vnet_interfaces(self.config)
        self.iproute.assert_called_once_with()

    def test_delete_vnet_interfaces_does_nothing_if_interfaces_do_not_exist(self):
        self.check_if_interface_exists.return_value = False
        delete_vnet_interfaces(self.config)
        self.assertFalse(self.iproute_obj.link.called)

    def test_delete_vnet_interfaces_check_if_interface_exists_for_each_interface_in_config(self):
        calls = [call("vnet-veth0"), call("vnet-br0"), call("vnet-br1")]
        delete_vnet_interfaces(self.config)
        self.check_if_interface_exists.assert_has_calls(calls)
        self.assertEqual(self.check_if_interface_exists.call_count, 3)

    def test_delete_vnet_interfaces_does_not_check_veth_interfaces_if_not_in_config(self):
        calls = [call("vnet-br0"), call("vnet-br1")]
        del self.config["veths"]
        delete_vnet_interfaces(self.config)
        self.check_if_interface_exists.assert_has_calls(calls)
        self.assertEqual(self.check_if_interface_exists.call_count, 2)

    def test_delete_vnet_interfaces_calls_ip_link_to_delete_interfaces(self):
        calls = [call("del", ifname=i) for i in ["vnet-veth0", "vnet-br0", "vnet-br1"]]
        delete_vnet_interfaces(self.config)
        self.iproute_obj.link.assert_has_calls(calls)
        self.assertEqual(self.iproute_obj.link.call_count, 3)

    def test_delete_vnet_interfaces_down_not_delete_veth_interfaces_if_not_in_config(self):
        calls = [call("del", ifname=i) for i in ["vnet-br0", "vnet-br1"]]
        del self.config["veths"]
        delete_vnet_interfaces(self.config)
        self.iproute_obj.link.assert_has_calls(calls)
        self.assertEqual(self.iproute_obj.link.call_count, 2)


class TestStartTcpdumpOnVNetInterface(VNetTestCase):
    def setUp(self) -> None:
        self.popen = self.set_up_patch("vnet_manager.operations.interface.Popen")

    def test_start_tcpdump_on_vnet_interface_makes_correct_popen_call(self):
        start_tcpdump_on_vnet_interface("dev1")
        self.popen.assert_called_once_with(
            shlex.split("tcpdump -i dev1 -U -w {}".format(join(settings.VNET_SNIFFER_PCAP_DIR, "dev1.pcap")))
        )



#!/usr/bin/env python3
import os
from setuptools import setup, find_packages

VERSION = "0.8.0"

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name="vnet-manager",
    version=VERSION,
    url="https://github.com/Erik-Lamers1/vnet-manager",
    download_url="https://github.com/Erik-Lamers1/vnet-manager/archive/{}.tar.gz".format(VERSION),
    packages=find_packages(exclude=["tools", "tools.*", "tests", "tests.*", "*.tests", "*.tests.*"]),
    author="Erik Lamers",
    license="MIT",
    # PyLXD 2.2.11 is currently broken: https://github.com/lxc/pylxd/issues/404
    install_requires=["colorama", "PyYAML", "tabulate", "pylxd==2.2.10", "pyroute2", "psutil", "distro"],
    entry_points={
        "console_scripts": [
            "vnet-manager = vnet_manager.vnet_manager:main",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



from io import StringIO
from unittest.mock import patch

from vnet_manager.tests import VNetTestCase
from vnet_manager.actions.help import display_help_for_action
from vnet_manager.conf import settings


class TestDisplayHelpForAction(VNetTestCase):
    def setUp(self) -> None:
        self.logger = self.set_up_patch("vnet_manager.actions.help.logger")

    def test_display_help_for_action_does_not_print_anything_if_unknown_action(self):
        display_help_for_action("banana")
        self.logger.warning.called_once_with("No help text available for action banana")
        self.assertFalse(self.logger.debug.called)

    @patch("sys.stdout", new_callable=StringIO)
    def test_display_help_for_action_prints_mapped_help_text(self, mock):
        display_help_for_action("start")
        self.assertEqual(mock.getvalue().strip(), settings.HELP_TEXT_ACTION_MAPPING["start"].strip())



from logging import getLogger

from vnet_manager.conf import settings

logger = getLogger(__name__)


def display_help_for_action(action: str) -> None:
    """
    Print a help text for the supplied action
    :param str action: The action to display the help text for
    """
    # First check if we have help text for this action
    if action not in settings.HELP_TEXT_ACTION_MAPPING:
        logger.warning("No help text available for action {}".format(action))
        return

    # Display the help text
    logger.debug("Displaying help text for action {}".format(action))
    print(settings.HELP_TEXT_ACTION_MAPPING[action])



from typing import Any, List, Union
import collections
import importlib
import os

ENV_VAR = "SETTINGS_MODULE"


class Settings(collections.abc.Mapping):
    def __init__(self, settings_module_name: str):
        mod = importlib.import_module(settings_module_name)

        # Assign al values to self
        for setting in filter(lambda x: x.isupper(), dir(mod)):
            setattr(self, setting, getattr(mod, setting))

    def __getitem__(self, item: str) -> Any:
        """Enable dict-like access to the settings"""
        try:
            return getattr(self, item)
        except AttributeError as e:
            raise KeyError(item) from e

    def __iter__(self) -> list:
        # return all upper case variables, but not internals
        return [x for x in dir(self) if x.isupper()]

    def __len__(self) -> int:
        return len(self.__iter__())


settings_module = os.environ.get(ENV_VAR, "vnet_manager.settings.base")

settings = Settings(settings_module)


def import_from_string(val: str, setting_name: str) -> Any:
    # Borrowed from https://github.com/tomchristie/django-rest-framework/blob/0dec36e/rest_framework/settings.py#L169
    """
    Attempt to import a class from a string representation.
    """
    try:
        # Nod to tastypie's use of importlib.
        parts = val.split(".")
        module_path, class_name = ".".join(parts[:-1]), parts[-1]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)
    except (ImportError, AttributeError) as e:
        msg = "Could not import '{}' for setting '{}'. {}: {}.".format(val, setting_name, e.__class__.__name__, e)
        raise ImportError(msg) from e


def perform_import(val: Any, setting_name: str) -> Union[Any, List[Any]]:
    # Borrowed from https://github.com/tomchristie/django-rest-framework/blob/0dec36e/rest_framework/settings.py#L155
    """
    If the given setting is a string import notation,
    then perform the necessary import or imports.
    """
    if isinstance(val, str):
        return import_from_string(val, setting_name)
    if isinstance(val, (list, tuple)):
        return [import_from_string(item, setting_name) for item in val]
    return val
